import logging
import numpy as np
import aug_sfutils as sf
from trview import red_time, time_slice

logger = logging.getLogger('trview.aug2icrf')


class ICRF:


    def __init__(self, nshot, tbeg=0., tend=10., dt_ds=0.001, icr_min=1e4, icr_diff=5e3):


        self.shot = nshot
        self.icr_min  = icr_min
        self.icr_diff = icr_diff
        diag = 'ICP'
        nant = 4
        self.freq  = np.zeros(nant)
        self.phase = np.zeros(nant)
        icp = sf.SFREAD(nshot, diag)
        if icp.status:
            pset_fr = icp.getparset('Frequenz')
            pset_ph = icp.getparset('AntPhase')
            for jant in range(nant):
                tbas = diag + '.'.ljust(9)
                sig1 = 'pnet%d'  %(jant + 1)
                sig2 = 'ploss%d' %(jant + 1)
                self.freq[jant] = pset_fr['Freq%d' %(jant+1)]

                logger.info('Requested signal %s', sig1)

                tim = icp.gettimebase(sig1)
                tbeg = max(tbeg, 1e-6)
                index = (tim >= tbeg) & (tim <= tend)
                self.t_icp = tim[index]
                ntim = len(self.t_icp)
                dat1 = icp.getobject(sig1)[index]
                tmp2 = icp.getobject(sig2)
                if tmp2 is None:
                    dat2 = np.zeros(ntim, dtype=np.float32)
                else:
                    dat2 = tmp2[index]

                if jant == 0:
                    self.picr = np.zeros((ntim, nant))

                self.picr[:, jant] = dat1 - dat2
            self.picr = np.maximum(self.picr, 0.)

# Downsample P_ICRF in time

            if dt_ds == 0.:
                self.t_ds    = self.t_icp
                self.picr_ds = self.picr
            else:
                logger.info('Downsampling PICRF(t) to %8.4f s', dt_ds)
                nt_ds = int((self.t_icp[-1] - self.t_icp[0] + 1e-3*dt_ds)/dt_ds)
                self.t_ds = self.t_icp[0] + 0.5*dt_ds + dt_ds*np.arange(nt_ds)
                self.picr_ds = time_slice.map2tgrid_1d(self.t_ds, self.t_icp, self.picr)

# Reduce time points, knots with significant changes

            ind_icr = red_time.red_time(icr_diff, self.picr)
            self.t_icr_red = self.t_icp[ind_icr]
            self.picr_red = self.picr[ind_icr, :]

            for jant in range(nant):
                if np.max(self.picr_red[:, jant]) <= icr_min:
                    self.picr_red[:, jant] = np.zeros(len(self.t_icr_red))

# Minority concentration

        diag = 'CXF'
        cxf = sf.SFREAD(nshot, diag)
        if cxf.status:
            sig = 'cHavR'
            if sig in cxf.sfh.keys():
                self.t_cxf = cxf.gettimebase(sig)
                self.minr = cxf.getobject(sig)
                self.minl = cxf.getobject('cHavL')


if __name__ == '__main__':

    ICRF(29795)
