import logging
import numpy as np
import aug_sfutils as sf
from trview import red_time, rot_matrix, time_slice

logger = logging.getLogger('trview.aug2nbi')

rot = rot_matrix.ROT_MATRIX

kv_def = {'INJ1': 60, 'INJ2': 93}


def dd2xy(inj_tor_box, src_slit, op, rabsca, phi_box):

    R_src = src_slit + op
    tmp   = rot(-inj_tor_box, R_src, rabsca, x_cen=op)
    alpha = np.arctan2(tmp.y, tmp.x)
    beta  = np.arctan2(tmp.y, tmp.x - op)
    rm = rot(phi_box, tmp.x, tmp.y)
    phi = alpha - beta
    return rm.x, rm.y, phi


def tr2as(rtcena, xybsca, xlbtna, xybapa, xlbapa, deltaR):

    rtcena    = np.atleast_1d(rtcena)
    xybsca    = np.atleast_1d(xybsca)
    xlbtna    = np.atleast_1d(xlbtna)
    xybapa    = np.atleast_1d(xybapa)
    xlbapa    = np.atleast_1d(xlbapa)

    vert_incl = -np.arcsin((xybapa - xybsca)/xlbapa)

    deltaR = np.atleast_1d(deltaR)
    rbtmin = 0.01*rtcena - 0.5*deltaR
    rbtmax = 0.01*rtcena + 0.5*deltaR
    hbeam  = 0.01*(xybsca - xlbtna*np.sin(vert_incl))
    tgA    = np.tan(vert_incl)

    return np.squeeze(rbtmin), np.squeeze(rbtmax), np.squeeze(hbeam), np.squeeze(tgA)


class NBI:


    def __init__(self, nshot, tbeg=0., tend=10., dt_ds=0., nbi_diff=1.e3, nbi_min=3e4):


        diag = 'NIS'
        sig = 'PNIQ'

        beta_deg = {}
        einj_nis = {}
        mass_amu = {}
        pow_mix  = {}
        nis = sf.SFREAD(nshot, diag)
        self.status = nis.status
        if nis.status:
            tim = nis.gettimebase(sig)
            dat = nis.getobject(sig)
            for box in ('INJ1', 'INJ2'):
                nips = nis.getparset(box)
                beta_deg[box] = nips['BETA']
                mass_amu[box] = nips['M']
                einj_nis[box] = nips['UEXQ']
                pow_mix [box] = nips['SPEC']
                ind = (einj_nis[box] <= 0)
                einj_nis[box][ind] = kv_def[box]
        else:
            logger.error('No NIS shotfile found for shot %d', nshot)
            return

# Parameters

        self.shot = nshot
        self.nbi_min  = nbi_min
        self.nbi_diff = nbi_diff
        self.n_box = np.array( [ len(einj_nis['INJ1']), len(einj_nis['INJ2']) ] )
        n_nbi = np.sum(self.n_box)
        self.nbi_on = np.ones(n_nbi, dtype=np.bool)

        if nshot > 14050:
            self.beta_sf_deg = np.append(beta_deg['INJ1'], beta_deg['INJ2'])
        else:
            self.beta_sf_deg = np.array(8*[4.8891] )
        abeam = np.append(np.repeat(mass_amu['INJ1'], self.n_box[0]), \
                          np.repeat(mass_amu['INJ2'], self.n_box[1]) )
        self.abeam = np.maximum(abeam, 1)
        self.zbeam = np.maximum(abeam/2, 1)
        self.einj_kev = np.append(einj_nis['INJ1'], einj_nis['INJ2'])

        self.power_mix = np.zeros((n_nbi, 3))
        self.part_mix  = np.zeros((n_nbi, 3))

        for jnb in range(n_nbi):
            if jnb == 0:
                powe  = pow_mix['INJ1'][::-1]
            elif jnb == self.n_box[0]:
                powe = pow_mix['INJ2'][::-1]
            pow_sum = np.sum(powe)
            if np.isnan(powe[0]) or pow_sum == 0:
                self.power_mix[jnb, :] = [1, 0, 0]
                self.part_mix [jnb, :] = [1, 0, 0]
                self.nbi_on[jnb] = False
            else:
                part = np.array([powe[0], 2.*powe[1], 3.*powe[2]])
                self.power_mix[jnb, :] = powe/pow_sum
                self.part_mix [jnb, :] = part/np.sum(part)

# Power

        index = (tim >= tbeg) & (tim <= tend)
        self.t_nis = tim[index]
        dat = dat[index]
        ntim = len(self.t_nis)
        self.pnbi = np.zeros((ntim, n_nbi))

        jnb = 0
        for jbox, box_size in enumerate(self.n_box):
            for jsrc in range(box_size):
                self.pnbi[:, jnb] = dat[:, jsrc, jbox]
                jnb += 1

# Downsample P_NBI in time

        if dt_ds == 0.:
            self.t_ds    = self.t_nis
            self.pnbi_ds = self.pnbi
        else:
            logger.info('Downsampling PNBI(t) to %8.4f s', dt_ds)
            nt_ds = int((self.t_nis[-1] - self.t_nis[0] + 1e-3*dt_ds)/dt_ds)
            self.t_ds = self.t_nis[0] + 0.5*dt_ds + dt_ds*np.arange(nt_ds)
            self.pnbi_ds = time_slice.map2tgrid_1d(self.t_ds, self.t_nis, self.pnbi)

# Reduce time points, knots with significant changes

        ind_nbi = red_time.red_time(nbi_diff, self.pnbi)
        nt_tra = len(ind_nbi)
        self.t_nbi_red = self.t_nis[ind_nbi]
        self.pnbi_red  = self.pnbi [ind_nbi, :]

        for jnb in range(n_nbi):
            if np.max(self.pnbi_red[:, jnb]) <= nbi_min:
                self.pnbi_red[:, jnb] = np.zeros(nt_tra)
                self.nbi_on[jnb] = False


    def aug_tr(self):

# distances: cm; angles: rad
# op: distance box-torus axis
# phi_box: absolute angle of NBI box (does not matter in poloidal cross section)
# inj_tor_box: toroidal angle between mid source and torus axis
# inj: toroidal angle between source and box aligned
# src_slit: distance ion source - aperture, for the center of the beamlines
# src_hw: horizontal semi-displacement of sources
# xybsca: source vertical elevation w.r.t. midplane
# rabsca: source horizontal half width w.r.t. mid-point
# RTCENA: R_tang
# vert_incl: vertical inclination
# XLBAPA: length = midpl1 / COS(vert_incl)
# midpl2: Grid - P(RTCENA)   projected on horizont. midplane
# XLBTNA: length = midpl2 / COS(vert_incl)
# XYBAPA: elevation at slit

# Rectangular half aperture in point P
      self.rapedga = 16
      self.xzpedga = 19

      self.phi_box = np.array(4*[np.radians(33.75)] + 4*[np.radians(209)])
      self.op = np.array(4*[284.20] + 4*[329.63])
      self.inj_tor_box = np.array(4*[np.radians(15)] + 4*[np.radians(18.9)])
      src_slit = 650.
      self.src_slit = np.array(8*[src_slit])
      self.src_hw  = 47.
      mag_ang = np.arctan2(self.src_hw, src_slit)
      self.midpl1  = np.hypot(self.src_slit, self.src_hw)

      self.magic_angle = np.array(2*[-mag_ang, mag_ang, mag_ang, -mag_ang])

      self.xybsca    = np.array([60, 60, -60, -60, 60, 70, -70, -60])
      self.rabsca    = self.src_hw*np.array(2*[1, -1, -1, 1])
      self.vert_incl = np.radians(self.beta_sf_deg)
      self.theta_los = np.array(2*[-1, -1, 1, 1])*self.vert_incl
      self.alpha     = self.inj_tor_box + self.magic_angle
      self.rtcena    = self.op*np.sin(self.alpha)
      self.midpl2    = self.midpl1 + np.sqrt(self.op**2 - self.rtcena**2)
      self.xlbtna    = self.midpl2/np.cos(self.vert_incl)
      self.xlbapa    = self.midpl1/np.cos(self.vert_incl)
      self.xybapa    = self.xybsca - np.sign(self.xybsca)*self.src_slit*np.tan(self.vert_incl)

      self.xsrc, self.ysrc, self.phi_los = dd2xy(self.inj_tor_box, self.src_slit, self.op, \
                            self.rabsca, self.phi_box)
      tmp = np.degrees(np.arctan2(self.ysrc, self.xsrc))
      ind = (self.phi_box > np.pi)
      tmp[ind] += 180
      self.xbzeta = tmp

# Species mix

      self.ffulla = self.part_mix[:, 0]
      self.fhalfa = self.part_mix[:, 1]
      self.ffulla[np.isnan(self.ffulla)] = 1.
      self.fhalfa[np.isnan(self.fhalfa)] = 0.
      self.einja  = 1e3*self.einj_kev


    def tr2as(self):

        if not hasattr(self, 'rtcena'):
            self.aug_tr()
        deltaR = [0.25, 0.35, 0.35, 0.25, 0.62, 0.58, 0.58, 0.62]
        self.aspect = 1
        self.rbtmin, self.rbtmax, self.hbeam, self.tgA = \
            tr2as(self.rtcena, self.xybsca, self.xlbtna, self.xybapa, self.xlbapa, deltaR)
        self.spec_arr = np.array([self.ffulla, 0.5*self.fhalfa, (1 - self.ffulla - self.fhalfa)/3.])
        self.spec_arr /= np.sum(self.spec_arr, axis=0) 


if __name__ == '__main__':

    NBI(28053)
