import logging
import numpy as np
import aug_sfutils as sf
from trview import CalcZef, time_slice

logger = logging.getLogger('trview.aug2sig')


diagsigs = {'ip': 'FPC:IpiFP', 'bt': 'FPC:BTF', 'uloop': 'TOT:u_loop' , \
    'prad': 'BPD:Pradtot', 'betpol': 'GQH:betpol' }


class READ:


    def __init__(self, equ, dt_ds=0.):

        nshot = equ.shot
        tbeg = equ.time[0]
        tend = equ.time[-1]
        for key, diagsig in diagsigs.items():
            tmp = type('', (), {})()
            diag, sig = diagsig.split(':')
            sfsig = sf.SFREAD(nshot, diag)
            if sfsig.status:
                tim = sfsig.gettimebase(sig)
                dat = sfsig.getobject(sig).copy()
                if dat is None:
                    if key in ('prad', 'betpol'):
                        logger.warning('Signal %s of diag %s not found', sig, diag)
                        continue

                (index, ) = np.where((tim >= tbeg) & (tim <= tend))
                if len(index) < 1:
                    index = np.argmin(np.abs(tim - tbeg))
                tmp.time = np.atleast_1d(tim[index])
                tmp.data = np.atleast_1d(dat[index])
                tmp.unit = dat.phys_unit
                tmp.diagsig = diagsig
                tmp.shot = nshot
                tmp.edition = sfsig.ed

# Downsample signal in time

                if dt_ds == 0.:
                    tmp.t_ds    = tmp.time
                    tmp.data_ds = tmp.data
                else:
                    logger.info('Downsampling signal %s:%s to %8.4f s', diag, sig, dt_ds)
                    nt_ds = int((tmp.time[-1] - tmp.time[0] + 1e-3*dt_ds)/dt_ds)
                    tmp.t_ds = tmp.time[0] + 0.5*dt_ds + dt_ds*np.arange(nt_ds)
                    tmp.data_ds = time_slice.map2tgrid_1d(tmp.t_ds, tmp.time, tmp.data)

                self.__dict__[key] = tmp
            else:
                logger.warning('Signal %s:%s not found for #%d', diag, sig, nshot)
                continue

#======
# Zef
#======

        try:
            tmp = type('', (), {})()

            zbrs = CalcZef.Zeff(equ)

            index = (zbrs.timeZeff >= tbeg) & (zbrs.timeZeff <= tend)
            tmp.time = np.atleast_1d(zbrs.timeZeff[index])
            tmp.data = np.atleast_1d(zbrs.Zeff0[index])
            tmp.unit = None
            tmp.shot = nshot
            tmp.diagsig = 'CX:Zef'
            tmp.edition = 0
            if np.max(np.abs(tmp.data)) == 0:
                tmp.data = zbrs.Zeff

        except:
            logger.error('CalcZef Bremsstrahlung unsuccesful')

        if not hasattr(self, 'zef') and nshot < 21482:
            try:
                tmp = type('', (), {})()

                zfml = CalcZef.Zeff_fml(nshot) # From formula
                index = (zfml.time >= tbeg) & (zfml.time <= tend)
                tmp.time = np.atleast_1d(zfml.time[index])
                tmp.data = np.atleast_1d(zfml.zeff[index])
                tmp.shot = nshot
                tmp.unit = None
                tmp.diagsig = 'HM:fml'
                tmp.edition = 0
                self.zef = tmp

            except:
                logger.exception('CalcZef formula unsuccesful')

        if hasattr(self, 'zef'):

# Downsample Zeff in time

            if dt_ds == 0.:
                tmp.t_ds    = tmp.time
                tmp.data_ds = tmp.data
            else:
                logger.info('Downsampling Zeff(t) to %d s', dt_ds)
                nt_ds = int((tmp.time[-1] - tmp.time[0] + 1e-3*dt_ds)/dt_ds)
                tmp.t_ds = tmp.time[0] + 0.5*dt_ds + dt_ds*np.arange(nt_ds)
                tmp.data_ds = time_slice.map2tgrid_1d(tmp.t_ds, tmp.time, tmp.data)

        else:

            tmp.time    = np.array([0, 10])
            tmp.t_ds    = np.array([0, 10])
            tmp.data    = np.array([1.4, 1.4])
            tmp.data_ds = np.array([1.4, 1.4])
            tmp.shot = nshot
            tmp.unit = None
            tmp.diagsig = 'CAR:onesix'
            tmp.edition = 0
            self.zef = tmp
