# Description: function for mapping of the irregulary sampled noise radial profiles
#              the regular grid. mapping is done by inversion of the grid projection 
#              matrix. Inversion is regularized by tikhonov regularization with 
#              an assumption of the minimal diffusion in a cylindrical geometry 
#		('power' density is expected as constant guess).  Moreover by 'zero_edge' 
#		can by profile forced to by zero at the edge,
#              Robusthes against outliers is improved by two pass calculation and 
    #               error correction by  Winsorizing estimator. 
# Author: Tomas Odstrcil
# mail:tomas.odstrcil@gmail.com

import os, sys, time, logging
from numpy import *
import gc
import scipy as sc
from matplotlib.pyplot import * 
import scipy.sparse as sp
import matplotlib.animation as manimation
from scipy.stats.mstats import mquantiles
from numpy.fft import fft, fftfreq, ifft, rfft2
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d

logger = logging.getLogger('trview.grid_map')


class map2grid():


    def __init__(self,R,T,Y,dy,nr_new,nt_new,eta=0,name=''):


        logger.info('grid map: %s', name)

        self.nr_new = nr_new
        self.nt_new = nt_new
        self.eta = eta
        self.name = name

        self.dtype='double'
        self.R = R.ravel()
        self.T = T.ravel()

        self.norm = linalg.norm(Y/amax(Y))/sqrt(size(Y))*amax(Y)

        Y/= self.norm    
        dy/= self.norm
        
        self.t_min = amin(self.T)
        self.t_max = amax(self.T)
        self.r_min = 0.
        self.r_max = max(min(amax(self.R),1.1),1.01)
    
    
        self.n_points = size(Y)
        self.Y = Y.ravel()
        self.dy = dy.ravel()
        self.dy[dy<=0] = infty
            
        self.g = zeros((nt_new,nr_new))*nan
        self.missing_data = ones(self.nt_new,dtype=bool)
        self.K = zeros((nr_new,nt_new))
        self.Kerr_u = zeros((nr_new,nt_new))*nan
        self.Kerr_d = zeros((nr_new,nt_new))*nan

        self.retro_f = zeros_like(self.Y)*nan
        
        self.chi2 = nan
        self.g1 = nan
        
        self.corrected=False
        self.prepared=False
        
        self.r_new,self.t_new = np.meshgrid(np.linspace(self.r_min,self.r_max,self.nr_new),
                np.linspace(self.t_min, self.t_max, self.nt_new))


    def FourierMatrix(self,n):
        #create matrix of the unitary fourier operator

        A = identity(n)
        F = fft(A)
        F/= sqrt(n/2)
        
        return F


    def G_filter(self,f, f0, s):
        #filter  of the higher orders of signal with frequency f0, width s

        f_resp = ones_like(f)
        if f0 is not None and f0 != 0:
            for i in range(1,int(amax(f)/f0)+2):
                f_resp*= 1-exp(-(f-i*f0*sign(f))**2/(2*(s/i**0.0)**2))/sqrt(1+((i-1)/5.)**2)
            
        #supress also higher harmonics 
        #higher  orders has a narrower window to force a higher stability 

        return f_resp

        
    def FilteredDerivative(self,n,f0,dt,missing_data):
       
        n_ext = 2**int(ceil(log2(n*2)))
        f = fftfreq(n_ext, d=dt)
        
        #speed of the nonperiodic changes
        s = 0.05
        GF = self.G_filter(f, f0, s)
        FT_D = 4*sin(pi*f*dt)**2

        F = self.FourierMatrix(n_ext)#fast
        
        spectrum = -FT_D*GF*sqrt(n_ext/2)

        D = ifft((spectrum[:,newaxis]*F).T)        
        D = real(D[:n,:n])  #use only a cut (D was 2x larger in order to avoid a edge effects)
         #imag part are only numerical errors
     
        D[missing_data,:] = (eye(n,k=1)-eye(n))[missing_data,:]
        
        #jumps = where(diff(int_(missing_data))==1)[0]+1
        #D[jumps,jumps] = 0
        #D[jumps,jumps+1] = 0
        #jumps = where(diff(int_(missing_data))==-1)[0]
        #D[jumps,jumps] = 0
        #D[jumps,jumps+1] = 0
     
        D-= diag(sum(D,axis=1))
        D[abs(D)<1e-5] = 0

        return D


    def _sparseness(self,x):
        """Hoyer's measure of sparsity for a vector"""

        sqrt_n = np.sqrt(len(x))
        return (sqrt_n - np.linalg.norm(x, 1) / np.linalg.norm(x)) / (sqrt_n - 1)


    def PrepareCalculation(self,eta=0,perturb_freq=0, zero_edge=False):

        logger.info('\nPrepareCalculation')
        
        #create a geometry matrix 
        dt = (self.t_max-self.t_min)/(self.nt_new-1)
        dr = (self.r_max-self.r_min)/(self.nr_new-1)
        it = (self.T-self.t_min)/dt
        it[it<0] = 0
        it[it>self.nt_new-1] = self.nt_new-1
        
        ir = (self.R-self.r_min)/dr
        ir[ir<0] = 0
        ir[ir>self.nr_new-1] = self.nr_new-1

        weight = empty((self.n_points,4))
        index_p = tile(arange(self.n_points,dtype=int), (4,1)).T
        index_t = empty((self.n_points,4),dtype=int)
        index_r = empty((self.n_points,4),dtype=int)

        index_t[:,0] = floor(it)
        index_t[:,1] = ceil(it)
        index_t[:,2] = index_t[:,0]
        index_t[:,3] = index_t[:,1]

        index_r[:,0] = floor(ir)
        index_r[:,1] = index_r[:,0]
        index_r[:,2] = ceil(ir)
        index_r[:,3] = index_r[:,2]

        weight[:,0] = (floor(it)+1-it)*(floor(ir)+1-ir)
        weight[:,1] = (it-floor(it))*(floor(ir)+1-ir)
        weight[:,2] = (floor(it)+1-it)*(ir-floor(ir))
        weight[:,3] = (it-floor(it))*(ir-floor(ir))
        
        #regions which are not covered by any measurements
        self.missing_data[index_t[:,0]] = False
        self.missing_data[index_t[:,1]] = False

        weight = weight.ravel()
        index_p = index_p.ravel()
        index_r = index_r.ravel()
        index_t = index_t.ravel()
        index_rt = index_r*self.nt_new+index_t
        npix = self.nr_new*self.nt_new
        
        self.vogel = sp.csc_matrix((copy(weight),(copy(index_p),copy(index_rt))),
                shape=(self.n_points,npix),dtype=self.dtype)
    
        #prepare smoothing matrix         
        #calculate (1+c)*d/dr(1/r*dF/dr) + (1-c)*d^2F/dt^2
        ct = exp(-eta)   #(0,1)
        cx = exp( eta)
        rvec = linspace(self.r_min,self.r_max,self.nr_new)
        rvec = (rvec[1:]+rvec[:-1])/2

        fun_r2 = (rvec*arctan(pi*rvec)-log((pi*rvec)**2+1)/(2*pi))/rvec  #alternative
        self.ifun =  1/sqrt(fun_r2)+1/(fun_r2)/4
        self.ifun =  1/fun_r2
        
        #time domain
        #dt = (t_max-t_min)/nt_new
        FD = self.FilteredDerivative(self.nt_new,perturb_freq,dt,self.missing_data)
        FD = sp.csc_matrix(FD,dtype=self.dtype)
    
        weight = r_[self.ifun.max(),self.ifun]*ct
    
        #radial domain
        DR = zeros((3,self.nr_new),dtype=self.dtype)
        
        DR[0,:-2] = self.ifun[:-1]
        DR[1,1:-1] = -self.ifun[:-1]-self.ifun[1:]
        DR[2,2:] =  self.ifun[1:]
        if zero_edge:
            DR[1,-1] = 10

        DR = sp.spdiags(DR*cx,(-1,0,1),self.nr_new,self.nr_new)
    
        #effective way to calculate (DR+DT).T*(DR+DT)
        W = sp.diags(weight,0)
        
        DDT = sp.kron(W**2,FD.T*FD)
        DRDT = sp.kron(DR.T*W,FD)+ sp.kron(W*DR,FD.T)
        DRDR = sp.kron(DR.T*DR,eye(self.nt_new), format='csc')
        
        DD = DDT+DRDT
        del DRDT, DDT
        DD = DD.tocsc()
        
        self.DD = DD+DRDR
        self.DT = sp.kron(sp.diags(weight,0),FD)
        self.DR = sp.kron(DR,eye(self.nt_new))
        del DRDR, FD

        self.prepared = True
            
        
    def PreCalculate(self):
        
        if not self.prepared:
            self.PrepareCalculation()
        
        logger.debug('First Pass')
        #first pass - decrease the weight of the outliers
        npix = self.nr_new*self.nt_new

        f = copy(self.Y)
        lam = 5

        f /= self.dy
        
        self.V = sp.spdiags(1/self.dy,0, self.n_points,self.n_points,format='csr')*self.vogel

        self.VV = self.V.T*self.V
            
        try:
            self.Factor = sp.linalg.factorized(self.VV+10**lam*self.DD)     
            g=squeeze(self.Factor(self.V.T*f))
        except :
            lam = lam+2
            self.Factor = sp.linalg.factorized(self.VV+10**lam*self.DD)
            g=squeeze(self.Factor(self.V.T*f))
       
        #make a more robust fit in one iteration
        dist = (self.V*g-self.Y/self.dy)  
        
        #Winsorizing M estimator
        dy_corr = ((dist/3)**2+1)**(0.25)  #BUG oprava

        self.f = f/dy_corr
        self.V = sp.spdiags(1/dy_corr,0, self.n_points,self.n_points,format='csr')*self.V
        self.VV = self.V.T*self.V
        
        self.corrected=True

        
    def Calculate(self,lam):

        if not self.corrected:
            self.PreCalculate()

        #noise vector for estimation of uncertainty
        n_noise_vec = 20
        noise = random.randn(self.n_points,n_noise_vec)
        noise /= ((noise/3)**2+1)**(0.25)
        
        try:
            self.Factor.cholesky_inplace(self.VV+10**lam*self.DD)
        except:
            self.Factor = sp.linalg.factorized(self.VV+10**lam*self.DD)

        g = squeeze(self.Factor(self.V.T*self.f))
        fg1 = log10(linalg.norm((self.vogel*g-self.Y)/self.dy)**2/sum(isfinite(self.dy)))

        map_noise = empty((self.nr_new*self.nt_new,n_noise_vec))

        #BUG
        for i in range(n_noise_vec):
            map_noise[:,i] = squeeze(self.Factor(self.V.T*noise[:,i]))

        self.chi2 = 10**fg1
        self.g1 = lam
        logger.debug('\nchi2: %.2f reg:%.2f' %(self.chi2, self.g1))

        map_noise = reshape(map_noise,(self.nr_new,self.nt_new,n_noise_vec)).T*self.norm*sqrt(self.chi2)
    
        retro_f = self.vogel*g
    
        self.retro_f = reshape(retro_f,shape(self.T))*self.norm    
        self.g       = reshape(g,(self.nr_new,self.nt_new)).T*self.norm
        
        if any(self.missing_data):
           self.g=interp1d(self.t_new[~self.missing_data,0],self.g[~self.missing_data,:],
                    axis=0,kind='linear',bounds_error=False, fill_value=nan)(self.t_new[:,0])

        K = -diff(self.g)*self.ifun[newaxis,:]*self.nr_new
        
        Kerr = std(diff(map_noise)*self.ifun[newaxis,newaxis,:],axis=0)*self.nr_new

        K_noise = diff(map_noise)*self.ifun[None,None,:]*self.nr_new
        K_noise.sort(0)
        K_noise-= mean(K_noise,axis=0)[newaxis,...]
        
        Kerr_u = abs(mean(K_noise[n_noise_vec//2:,...],axis=0))/0.76 #84% quantil
        Kerr_d = abs(mean(K_noise[:n_noise_vec//2,...],axis=0))/0.76 #16% quantil

        K0 = zeros((1,self.nt_new))
        self.K,self.Kerr_u,self.Kerr_d=r_[K.T,K0*nan],r_[Kerr_u.T,K0],r_[Kerr_d.T,K0]
        self.Kerr = r_[Kerr.T,K0]

        return self.g, self.t_new, self.r_new
        
        
    def PlotFit(self,show_plot,make_movie):
        
        path = os.getenv('HOME')+'/tr_client/profiles/'
        if not os.path.exists(path): os.makedirs(path)
        
        fig = figure('retro 2',figsize=(18,10))
        fig.clf()
        
        self.Y  = reshape(self.Y ,shape(self.T))*self.norm
        self.dy = reshape(self.dy,shape(self.T))*self.norm
        T = copy(self.T)
        T[T > amax(T)-1e-2] = nan
        self.Y[self.Y == 0] = nan
        ax = fig.add_subplot(111)
        ax.plot(T,self.retro_f,'r-')   
        ax.errorbar(T.ravel(),self.Y.ravel(),
                yerr=self.dy.ravel(),lw=0.1,capsize=0,fmt='-')
        ax.axis('tight')
        
        maxlim = mquantiles(self.retro_f,0.99)
        minlim = mquantiles(self.retro_f,0.02)
        if minlim == maxlim:  minlim,maxlim = 0,1
        ax.set_ylim(minlim-(maxlim-minlim)*0.1,maxlim+(maxlim-minlim)*0.1)
        
        fig.savefig(path+self.name+'retrofit_t.png')
        
        fig = figure('smooth fit2')
        fig.clf()

        ax = fig.add_subplot(111)
        ax.plot(self.t_new[:,1:],self.K.T,'k-',lw=0.2) 
        ax.axis('tight')
        fig.show()
        
        fig.savefig(path+self.name+'fit_t.png')
    
        if not (show_plot or make_movie):
            return self.g, self.t_new, self.r_new
# Plots
        extent = [self.r_min,self.r_max ,self.t_min,self.t_max]
    
        fig1 = figure('invert diffusion')
        fig1.clf()
        ax_im1 = fig1.add_subplot(111)
        im = ax_im1.imshow(self.K[::-1,:].T,extent=extent, aspect='auto')
        #ax_im1.axis(extent)
        ax_im1.set_xlabel('$\\rho_{pol}$ [-]')
        ax_im1.set_ylabel('$t$ [s]')
        
        fig1.colorbar(im)
        CS = ax_im1.contour(self.K.T,[0,],c='w',origin='lower',lw=1,
                extent=extent)
        fig1.savefig(path+self.name+'_diffusion.png')
        pause(0.01)
    
        fig2 = figure('smooth fit')
        fig2.clf()
        ax_im2 = fig2.add_subplot(111)
        im = ax_im2.imshow(self.g[::-1,:], extent=extent, aspect='auto')
        ax_im2.axis(extent)

        ax_im2.set_xlabel('$\\rho_{pol}$ [-]')
        ax_im2.set_ylabel('$t$ [s]')
        fig2.colorbar(im)

        CS = ax_im2.contour(self.g,[0,],c='w',origin='lower',lw=1,extent=extent)
        fig2.savefig(path+self.name+'_fit.png')
    
        pause(0.01)
 
        if not (show_plot or make_movie):
            return self.g, self.t_new, self.r_new

        fig = figure('retrofit',figsize=(7,8))
        fig.clf()
        ax1 = fig.add_subplot(211)
        replot_plot, = ax1.plot([],[],'+',label='retrofit', animated=not make_movie,zorder=98)
        plotline, caplines, barlinecols = ax1.errorbar([],[],[],fmt='.',
                        label='measured data', animated=not make_movie)

        fit_plot, = ax1.plot([],[],'-',linewidth = 0.5, animated=not make_movie,zorder=99)
        maxlim = mquantiles(self.g,0.98)
        minlim = mquantiles(self.g,0.02)
        if minlim == maxlim:  minlim,maxlim = 0,1
        ax1.set_ylim(minlim-(maxlim-minlim)*0.1,maxlim+(maxlim-minlim)*0.3)

        ax1.set_xlim(0,1.05)
        ax1.legend()
        ax1.xaxis.set_major_formatter( NullFormatter() )
        ax1.set_title('Fitted data: '+self.name+',   '+'chi2: %.2f reg:%.2f'%(self.chi2,self.g1))
        
        title_template = '%s, time: %.3fs'    # prints running simulation time
        time_text = ax1.text(0.05, 0.9, '', transform=ax1.transAxes)
        
        ax2 = fig.add_subplot(212)
        ax2.set_title('Numerical inverted diffusion')
        k_plot,= ax2.plot([],[],'r-', animated=not make_movie)
        confidence_plot = ax2.fill_between(self.r_new[0,:-1],ones(self.nr_new-1)*1.5,
                        ones(self.nr_new-1), alpha=.2, facecolor='b', edgecolor='None',
                        label='68% confidence interval', animated=not make_movie)
        ax2.set_xlim(0,1.05)
        maxlim = mquantiles(self.K[:-1,:],0.98)
        minlim = min(0,mquantiles(self.K[:-1,:],0.02))

        if minlim == maxlim:  minlim,maxlim = 0,1
        ax2.set_ylim(minlim-(maxlim-minlim)*0.1,maxlim+(maxlim-minlim)*0.1)
        ax2.set_xlabel('$\\rho_{pol}$ [-]')
        ax2.axhline(y=0, linestyle='--')
        
        FFMpegWriter = manimation.writers['mencoder']
        metadata = dict(title=self.name, artist='Matplotlib', comment='profile animation')
        writer = FFMpegWriter(fps=20, metadata=metadata,codec='mpeg4',bitrate=2000)

        dt_step = (self.t_max-self.t_min)/self.nt_new
        
        run_time = time.time()
        with writer.saving(fig, path+self.name+'_profile.wmv', 100):
            for t in linspace(self.t_min,self.t_max,self.nt_new):  

                fps = 1/(time.time()-run_time)
                run_time = time.time()

                if make_movie:
                    sys.stdout.write('\r'+path+self.name+'_profile.wmv <-'+" writting %2.3fs fps:%2.1f"%(t,fps))
                else:
                    sys.stdout.write('\r'+self.name+"  plotting %2.3fs, fps: %2.1f"%(t,fps))
   
                sys.stdout.flush()

                ind = (abs(self.T- t) < dt_step/2) 
                
                time_text.set_text('time: %.3fs'%t)
                
                if any(ind):

                    x = self.R[ind]
                    y = self.Y[ind]
                    yerr = self.dy[ind]
                    ry = self.retro_f[ind]
                    replot_plot.set_data(x,ry)

                    # Replot the data first
                    plotline.set_data(x,y)

                    # Find the ending points of the errorbars
                    error_positions = (x,y-yerr), (x,y+yerr)

                    # Update the caplines
                    for j,pos in enumerate(error_positions):
                        caplines[j].set_data(pos)

                    # Update the error bars
                    barlinecols[0].set_segments(zip(zip(x,y-yerr), zip(x,y+yerr))) 
                    
                ind = argmin(abs(self.t_new[:,0]-t))
                fit_plot.set_data(self.r_new[ind,:].T,self.g[ind,:].T)

                paths, = confidence_plot.get_paths()
                
                #inv K and errorbars
                k_plot.set_data(self.r_new[ind,:].T,self.K[:,ind])
            
                u_err = self.K[:-1,ind]+self.Kerr_u[:-1,ind]
                b_err = self.K[:-1,ind]-self.Kerr_d[:-1,ind]

                paths.vertices[1:self.nr_new,1] = u_err

                paths.vertices[self.nr_new+1:-1,1] = b_err[::-1]
                paths.vertices[self.nr_new,1] = b_err[-1]
                paths.vertices[0,1] = b_err[0]#0.2
            
                writer.grab_frame() 

        close()

        return

        
    def finish(self):

        try:
            del self.Factor, self.DD,self.DT,self.DR,self.VV, self.V
            gc.collect()
        except:
            pass


    def OptimalReg(self):
           
        res = 20
        L = linspace(0,6,res)
        g = zeros((self.nr_new*self.nt_new,res))
        CH = zeros(res)
        S1 = zeros(res)
        S2 = zeros(res)
        S3 = zeros(res)
        S4 = zeros(res)
        A = zeros(res)
        DoF = sum(isfinite(self.dy))
        
        for i,l in enumerate(L):
            sys.stdout.write('\rcalc. L-curve %2.1f%%'%((i*100.)/res))   
            sys.stdout.flush()
            t = time.time()
            
            try:
                self.Factor = sp.linalg.factorized(self.VV+10**l*self.DD)
                g[:,i] = squeeze(self.Factor(self.V.T*self.f))

            except:
                continue
            
            CH[i] = log10(linalg.norm((self.vogel*g[:,i]-self.Y)/self.dy)**2/DoF)
            S1[i] = log10(linalg.norm((self.DT+self.DR)*g[:,i]))
            S2[i] = log10(linalg.norm(self.DR*g[:,i]))
            S3[i] = log10(linalg.norm(self.DT*g[:,i]))
            clf()
            subplot(121)
            
            K = -(diff(reshape(g[:,i],(self.nr_new,self.nt_new)),axis=0)*self.ifun[:,None])
     
            D = diff(K,1,axis=0)
            FD = rfft2(D-mean(D))

            A[i] = 1/(1-self._sparseness(abs(FD).ravel()))
        
        neg_ind = diff(r_[A,-infty]) < 0
        i_neg = where(diff(int_(neg_ind))==1)[0][-1]
        ind = (A>2)& (A<7)&(range(res)<i_neg)

        figL = figure('L-curve')
        figL.clf()

        figL.subplots_adjust(hspace=0, wspace = 0)
        ax = figL.add_subplot(211)
        ax.set_ylabel('$\log_{10}||\Gamma x/\lambda||$')
        ax.xaxis.set_major_formatter( NullFormatter() )

        L_inter = linspace(amin(L),amax(L),1000)
        C = diff(InterpolatedUnivariateSpline(L,A,k=4)(L_inter))/diff(L_inter)
        As = InterpolatedUnivariateSpline(L,A,k=4)(L_inter)

        X = InterpolatedUnivariateSpline(L,CH,k=4)(L_inter)
        Cch = diff(InterpolatedUnivariateSpline(L,A,k=4)(L_inter))/diff(X)
        X = X[1:]
        
        fig = figure('L')
        fig.clf()
        ax = fig.add_subplot(111)
        
        if not any(ind):
            i = argmax((diff(A)/diff(CH)))
        else:
            i = where(ind)[0][argmax((diff(A)/diff(CH))[ind[:-1]])]

        ax.plot(X,C)
        ax.plot(X,Cch/20)

        ax.plot((CH[1:]+CH[:-1])/2, diff(A)/diff(L), 'o')
        ax.plot((CH[1:]+CH[:-1])/2, diff(A)/diff(CH)/20, '+')

        ax.axvline((CH[1:]+CH[:-1])[i]/2)
        ax.plot(CH[ind], A[ind], 'bx')
        ax.plot(CH[~ind], A[~ind], 'rx')

        ax.plot(X, InterpolatedUnivariateSpline(L,A,k=4)(L_inter)[1:])
        ymin,ymax = ax.get_ylim()
        ax.set_ylim(max(-10,ymin),min(20,ymax))
        
        logger.debug(L[i], i)

        return L[i]
