import os, logging
import numpy as np
try:
    import Tkinter as tk
    import ttk
except:
    import tkinter as tk
    from tkinter import ttk

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
try:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk as nt2tk
except:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg as nt2tk
from matplotlib.figure import Figure
from matplotlib.widgets import Slider
from aug_sfutils import getgc
from trview.tooltip import createToolTip

logger = logging.getLogger('trview.plot_geo')

figc = '#70a0c0' # Figure


class GEO:


    def __init__(self, fit_sep, geoframe=None):

        self.fit = fit_sep

        if geoframe is None:
            if __name__ == '__main__':
                geoframe = tk.Tk()
                import trgui_style
            else:
                geoframe = tk.Toplevel()
        xmax = geoframe.winfo_screenwidth()
        ymax = geoframe.winfo_screenheight()
        width  = min(1200, int(0.95*xmax))
        height = min(800, int(0.88*ymax)) 

        pol_frame = ttk.Frame(geoframe)
        butframe  = ttk.Frame(geoframe)
        toolframe = ttk.Frame(geoframe)
        for frame in (toolframe, butframe, pol_frame):
            frame.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=1)
 
# Canvas frame
        fig_dpi = 100
        fig_hei = height - 120
        fig_size = fig_hei/fig_dpi
        self.pol_fig = Figure(dpi=fig_dpi)
        self.pol_fig.set_size_inches(fig_size, fig_size)#, forward=True)
        self.pol_fig.set_facecolor(figc)
        self.pol_fig.subplots_adjust(left=0.1, bottom=0.1, right=0.9 , top=0.94)
        self.pol_canvas = FigureCanvasTkAgg(self.pol_fig, master=pol_frame)
        self.pol_canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        axsep = self.pol_fig.add_subplot(111, aspect='equal')
        axsep.set_xlim((0.8, 3.0))
        axsep.set_ylim((-1.4, 1.4))
        axsep.set_xlabel('R [m]')
        axsep.set_ylabel('z [m]')
        self.septxt = axsep.text(2.6, -1.1, 't = 0.000 s', ha='center')
        self.errtxt = axsep.text(2.6, .7, '', ha='center')

        gc_d = getgc()
        for gc in gc_d.values():
            axsep.plot(gc.r, gc.z, 'b-')

        self.sepscat, = axsep.plot([], [], 'k+')
        self.mag,     = axsep.plot([], [], 'ro', label='Mag. axis')
        self.u,       = axsep.plot([], [], 'ko')
        self.o,       = axsep.plot([], [], 'go')
        self.xu,      = axsep.plot([], [], 'bo', label='Lower X')
        self.xo,      = axsep.plot([], [], 'mo', label='Upper X')
        self.sepfit,  = axsep.plot([], [], 'g-')
        axsep.legend(numpoints=1, loc=1)

# Player buttons

        locdir = os.path.dirname(os.path.realpath(__file__))
        self.playfig     = tk.PhotoImage(file='%s/play.gif'     %locdir)
        self.pausefig    = tk.PhotoImage(file='%s/pause.gif'    %locdir)
        self.forwardfig  = tk.PhotoImage(file='%s/forward.gif'  %locdir)
        self.backwardfig = tk.PhotoImage(file='%s/backward.gif' %locdir)

        self.play_button = ttk.Button(butframe, command=self.Play,    image=self.playfig)
        backward_button = ttk.Button(butframe, command=self.Backward, image=self.backwardfig)
        forward_button  = ttk.Button(butframe, command=self.Forward,  image=self.forwardfig)

        createToolTip(forward_button  , 'Go to next timestep')
        createToolTip(backward_button , 'Go backward by one timestep')
        createToolTip(self.play_button, 'Forward animation/pause')
        try:
            sliderax = self.pol_fig.add_axes([0.1, 0.01, 0.6, 0.03], facecolor='yellow')
        except:
            sliderax = self.pol_fig.add_axes([0.1, 0.01, 0.6, 0.03], axisbg='yellow')
        self.sl_time = Slider(sliderax,'Time:', 0., 1., valinit=0.)
        self.sl_time.valtext.set_text('0.0000')

        def update_time(val):
            self.update_plot(val=val)

        self.sl_time.on_changed(update_time)
        self.pol_fig.canvas.mpl_connect('button_press_event', lambda event:self.pol_fig.canvas._tkcanvas.focus_set())

        for but in backward_button, self.play_button, forward_button:
            but.pack(side=tk.LEFT)

# Toolbar

        toolbar = nt2tk(self.pol_canvas, toolframe)
        toolbar.update()

        self.pol_canvas.draw()
        if __name__ == '__main__':
            geoframe.mainloop()

# Plot

        self.nt = len(self.fit.time)

        axsep.set_title('%s:%s(%d)  #%d' \
                    %(self.fit.exp, self.fit.diag, self.fit.ed, self.fit.shot) )

        self.begin()

#--------------------------
# Animation separatrix plot
#--------------------------

    def begin(self):
        self.stop  = True
        self.jt = 0
        self.update_plot()

    def Pause(self):
        self.stop = True
        self.play_button['image']   = self.playfig
        self.play_button['command'] = self.Play

    def Forward(self):
        self.stop = True
        self.play_button['image']   = self.playfig
        self.play_button['command'] = self.Play
        self.jt += 1
        self.jt = self.jt%self.nt
        self.update_plot()

    def Backward(self):
        if not hasattr(self, 'jt'):
            logger.error('Run\n    File-> Read and fit separatrix\nfirst')
            return
        self.stop = True
        self.play_button['image']   = self.playfig
        self.play_button['command'] = self.Play
        self.jt -= 1
        self.jt = self.jt%self.nt
        self.update_plot()

    def Play(self):
        self.stop = False
        self.play_button['image']   = self.pausefig
        self.play_button['command'] = self.Pause

        timeout = 1e-10
        while (self.jt <  self.nt-1) and (not self.stop):
            self.jt += 1
            self.update_plot()
            self.pol_canvas.start_event_loop(timeout)


    def updateTimeSlider(self, val=None):
        if val is None:
            progress = (self.fit.time[self.jt] - self.fit.time[0])/(self.fit.time[-1] - self.fit.time[0])
        else:
            progress = val
        self.sl_time.val = progress
        poly = self.sl_time.poly.get_xy()
        poly[2:4, 0] = progress
        self.sl_time.poly.set_xy(poly)
        self.sl_time.valtext.set_text('%8.4f' %self.fit.time[self.jt])


    def update_plot(self, val=None):

        if val is not None:
            tloc = self.fit.time[0] + val*(self.fit.time[-1] - self.fit.time[0])
            self.jt = np.argmin(np.abs(self.fit.time - tloc))

        if not hasattr(self, 'fit'):
            return
        self.sepscat.set_data(self.fit.rscat[self.jt], self.fit.zscat[self.jt])

        if self.fit.error[self.jt] == 0:
            self.sepfit.set_visible(True)
            self.sepfit.set_data(self.fit.r_bnd[self.jt], self.fit.z_bnd[self.jt])
            self.errtxt.set_text('')
        else:
            self.sepfit.set_visible(False)
            if self.fit.error[self.jt] == 1:
                self.errtxt.set_text('Close to ELM')
            elif self.fit.error[self.jt] == 2:
                self.errtxt.set_text('max R > 2.2')
            elif self.fit.error[self.jt] == 3:
                self.errtxt.set_text('min z < -1.1')
            elif self.fit.error[self.jt] == 4:
                self.errtxt.set_text('max z > 1')
            elif self.fit.error[self.jt] == 5:
                self.errtxt.set_text('All moments <= 0')
            elif self.fit.error[self.jt] == 6:
                self.errtxt.set_text('min z-fit < -1.1')

        if hasattr(self.fit, 'Rmag'):
            self.mag.set_data(self.fit.Rmag[self.jt], self.fit.Zmag[self.jt])
            self.u.set_data( self.fit.Rzunt[self.jt], self.fit.Zunt[self.jt])
            self.o.set_data(self.fit.Rzoben[self.jt], self.fit.Zoben[self.jt])
            self.xu.set_data( self.fit.Rxpu[self.jt], self.fit.Zxpu[self.jt])
            self.xo.set_data( self.fit.Rxpo[self.jt], self.fit.Zxpo[self.jt])
        self.septxt.set_text('t =%6.3f s' %self.fit.time[self.jt])
        self.updateTimeSlider(val=val)
        self.pol_canvas.draw()


if __name__ == '__main__':
# Execute fit

    nshot = 28053
    tbeg = 1.
    tend = 3.
    exp  = 'AUGD'
    diag = 'EQI'
    ed   = 0
    nmom = 6
    n_the = 101
    fit_sep = fit_sep.FIT_SEP(nshot, tbeg, tend, diag=diag, exp=exp, ed=ed, mom_order=nmom, n_the=n_the)

    GEO(fit_sep)
