import os, sys, time, logging
try:
    import Tkinter as tk
    import ttk
except:
    import tkinter as tk
    from tkinter import ttk
import numpy as np
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
try:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk as nt2tk
except:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg as nt2tk
from matplotlib.figure import Figure
import matplotlib.pylab as plt
import mpl_toolkits.mplot3d.axes3d as p3
from scipy.stats.mstats import mquantiles
from matplotlib.widgets import Slider
from trview.grid_map import map2grid
from trview.tooltip import createToolTip

logger = logging.getLogger('trview.tkmpl')


class DataFit():


    def __init__(self, nshot, n_spl, rdd, eq_exp='AUGD', eq_diag='EQH', eq_ed=0, myframe=None):


        if hasattr(self, 'can_frame'):
            self.can_frame.destroy()
        if myframe is None:
            myframe = tk.Toplevel()
            myframe.title('View and edit profiles')
            myframe.geometry('1200x740')


# Prepare + merge data

        self.ext_str = ''
        self.ylab    = '' 
        self.sflists = sorted(rdd.keys())
        for sflist in self.sflists:
            diag = sflist.split(':')[1]
            self.ext_str += diag
            self.ylab    += ',%s' %diag
        self.ylab = self.ylab[1:]

        self.nshot = nshot
        self.nr_fit = n_spl

        self.eq_ed   = eq_ed
        self.eq_diag = eq_diag
        self.eq_exp  = eq_exp

        for diag_sig in rdd.keys():
            if not hasattr(self, 'xexp'):
                self.xexp = np.array(rdd[diag_sig]['rhop'])
                self.yexp = np.array(rdd[diag_sig]['data'])
                self.yerr = np.array(rdd[diag_sig]['data_err'])
            else:
                self.xexp = np.append(self.xexp, rdd[diag_sig]['rhop']    , axis=1)
                self.yexp = np.append(self.yexp, rdd[diag_sig]['data']    , axis=1)
                self.yerr = np.append(self.yerr, rdd[diag_sig]['data_err'], axis=1)

        self.tgrid = rdd[diag_sig]['tgrid']

        self.perturb_freq = 0

        self.channel = range(self.xexp.shape[1])

        tbeg = np.min(self.tgrid)
        tend = np.max(self.tgrid)

        self.jt = 0

        self.eta = 0
        self.lam = 4

#        dt = self.tgrid[1] - self.tgrid[0]
#        self.nt = int((tend - tbeg)/dt) + 1
        self.nt = len(self.tgrid)

        self.xlab = r'$\rho_{pol}$'

        rho  = self.xexp.ravel()
        y    = self.yexp.ravel()
        yerr = self.yerr.ravel()
        tvec = np.repeat(self.tgrid, self.xexp.shape[1])

# Create object of the fitting routine
        self.m2g = map2grid(rho, tvec, y, yerr, self.nr_fit, self.nt, \
            name=self.ext_str)

        self.fit_prepared = False

# Canvas frame
        self.fig = Figure(figsize=(7., 4.), dpi=100)
        self.ax1 = self.fig.add_subplot(1, 2, 1)
        self.ax2 = self.fig.add_subplot(1, 2, 2)

        self.can_frame = ttk.Frame(myframe)
        self.can_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        self.canvas = FigureCanvasTkAgg(self.fig, master=self.can_frame)
        toolbar = nt2tk(self.canvas, myframe)
        toolbar.update()

        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)
        self.canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        fsize = 12

# Plots in the frame
        self.plotline, self.caplines, self.barlinecols=self.ax1.errorbar( \
                                                                          0,0,0, fmt='g.', label='measured data', capsize=3.)
        self.replot_plot, = self.ax1.plot([], [], 'b+', label='retrofit')
        self.fit_plot,    = self.ax1.plot([], [], 'k-', linewidth=.5)

        self.ax1.set_xlim(0, 1.05)
        self.ax1.legend()
        self.ax1.xaxis.set_major_formatter( plt.NullFormatter() )

        title_template = '%s, time: %.3fs'    # prints running simulation time
        self.time_text = self.ax1.text(.05, .9, '', transform=self.ax1.transAxes)

        self.ax2.set_title('Numerical inverted diffusion')
        self.k_plot, = self.ax2.plot([], [], 'r-', label='$1/\chi_N$')
        self.confidence_plot = self.ax2.fill_between(self.m2g.r_new[0, :-1], \
            np.ones(self.nr_fit-1)*1.5, np.ones(self.nr_fit-1), \
            alpha=.2, facecolor='b', edgecolor='None', \
            label='68% confidence interval')
        self.ax2.set_xlim(0, 1.05)

        self.ax2.set_xlabel(self.xlab, fontsize=fsize)
        self.ax1.set_xlabel(self.xlab, fontsize=fsize)
        self.ax1.set_ylabel(self.ylab, fontsize=fsize)

        hbox1 = tk.Frame(myframe)
        hbox2 = tk.Frame(myframe)
        hbox3 = tk.Frame(myframe)
        for box in hbox1, hbox2, hbox3:
            box.pack(side=tk.TOP, fill=tk.X)
# Player buttons

        locdir = os.path.dirname(os.path.realpath(__file__))
        self.playfig  = tk.PhotoImage(file='%s/play.gif'     %locdir)
        self.pausefig = tk.PhotoImage(file='%s/pause.gif'    %locdir)
        self.forwardfig  = tk.PhotoImage(file='%s/forward.gif'  %locdir)
        self.backwardfig = tk.PhotoImage(file='%s/backward.gif' %locdir)

        backward_button  = ttk.Button(hbox1, command=self.Backward, image=self.backwardfig)
        self.play_button = ttk.Button(hbox1, command=self.Play    , image=self.playfig)
        forward_button   = ttk.Button(hbox1, command=self.Forward , image=self.forwardfig)
        backward_button.pack( side=tk.LEFT)
        self.play_button.pack(side=tk.LEFT)
        forward_button.pack(  side=tk.LEFT)

# Commands

        btfit = tk.Button(hbox2, text='Re-fit', command=self.fit)
        btplo = tk.Button(hbox2, text='Plot 3D', command=self.plot3d)
        for but in btfit, btplo:
            but.pack(side=tk.LEFT)

# User instructions

        mouse_help = tk.Label(hbox3, text='Mouse: ')
        mouse_left = tk.Label(hbox3, text='Left (+Ctrl): del point (Channel)  ', \
                              fg="#900000")
        mouse_mid   = tk.Label(hbox3, text='Mid: re-fit ', fg="#009000")
        mouse_right = tk.Label(hbox3, text='Right: next ', fg="#000090")

        for w in (mouse_help, mouse_left, mouse_mid, mouse_right):
            w.pack(side=tk.LEFT)

        self.stop  = True
        self.ctrl  = False
        self.shift = False

        def stop_handler(event=None, self=self):
            self.stop  = True

        forward_button.bind( "<Button-1>", stop_handler)
        backward_button.bind("<Button-1>", stop_handler)

        self.entr_per_freq = tk.Entry(hbox2, width=2)
        self.entr_per_freq.insert(0, self.perturb_freq)
        self.entr_per_freq.pack(side=tk.RIGHT)
        cfit_lbl = tk.Label(hbox2, text='Perturbation freq. [Hz]')
        cfit_lbl.pack(side=tk.RIGHT)

        axcolor = 'lightgoldenrodyellow'

        self.fig.subplots_adjust(left=.15, bottom=.25, right=.95, top=.95, \
                                 hspace=.1, wspace=0)
        try: #matplotlib > 2.0
            self.sl_ax_time = self.fig.add_axes([.2, .12, .7, .05], facecolor=axcolor)
            sl_ax  = self.fig.add_axes([.2, .03, .3, .05], facecolor=axcolor)
            sl_ax2 = self.fig.add_axes([.6, .03, .3, .05], facecolor=axcolor)
        except:
            self.sl_ax_time = self.fig.add_axes([.2, .12, .7, .05], axisbg=axcolor)
            sl_ax  = self.fig.add_axes([.2, .03, .3, .05], axisbg=axcolor)
            sl_ax2 = self.fig.add_axes([.6, .03, .3, .05], axisbg=axcolor)
        self.sl_time = Slider(self.sl_ax_time, 'Time:', tbeg, tend, valinit=tbeg)
        self.sl_eta = Slider(sl_ax , '', -3, 3, valinit=self.eta)
        self.sl_lam = Slider(sl_ax2, '', 0, 7, valinit=self.lam)

        self.fig.text(.2, .095, 'Time smoothing -->:')
        self.fig.text(.6, .095, 'Total smoothing -->:')

        createToolTip(forward_button, 'Go forward by one timestep')
        createToolTip(backward_button, 'Go backward by one timestep')
        createToolTip(self.play_button, 'Go step by step forward, pause by second press')
        createToolTip(self.entr_per_freq, 'Frequency of the periodic plasma perturbations (it is slow!).')

        def update_eta(val):
            self.fit_prepared=False
            stop_handler()

        def update_lam(val):
            stop_handler()

        def update_time(val):
            self.jt = int((val - self.tgrid[0])/(self.tgrid[1] - self.tgrid[0]) + 0.5)
            self.plot_step()
            self.updateTimeSlider(val)
            
        self.sl_time.on_changed(update_time)

        self.fig.canvas.mpl_connect('button_press_event', self.MouseInteraction)
        self.fig.canvas.mpl_connect('key_press_event',   self.on_key)
        self.fig.canvas.mpl_connect('key_release_event', self.off_key)
        self.fig.canvas.mpl_connect('button_press_event', lambda event:self.fig.canvas._tkcanvas.focus_set())

        self.sl_eta.on_changed(update_eta)
        self.sl_lam.on_changed(update_lam)

        self.PreparePloting()
        self.fit()


    def fit(self):
# Perform fit

        eta = self.sl_eta.val
        f0 = float(self.entr_per_freq.get().strip())
        edge = False

        if not self.fit_prepared or f0!= self.perturb_freq:
            self.m2g.PrepareCalculation(eta=-eta, perturb_freq=f0, zero_edge=edge)
            self.fit_prepared = True
            self.perturb_freq = f0

        g, t_out, x_out = self.m2g.Calculate(self.sl_lam.val)

        self.tnot = []
        self.x_out = self.m2g.r_new[0, :]
        self.yfit_rhop = self.m2g.g
        self.fit_tol = self.sl_eta.val
        self.sigmas = self.sl_lam.val
        self.method = '2D regularisation'

        self.PreparePloting()


    def Pause(self):
        self.stop = True
        self.play_button['image']=self.playfig
        self.play_button['command']=self.Play


    def Forward(self):
        self.jt += 1
        self.plot_step()
        self.updateTimeSlider(self.tgrid[self.jt])


    def Backward(self):
        self.jt -= 1
        self.plot_step()
        self.updateTimeSlider(self.tgrid[self.jt])


    def Play(self):

        self.stop = False
        self.play_button['image']=self.pausefig
        self.play_button['command']=self.Pause

        while (self.jt < self.nt):

            if self.stop:
                break

            self.can_frame.after(1, self.plot_step)
            time.sleep(1e-3)

            self.canvas.get_tk_widget().update()

            self.updateTimeSlider(self.tgrid[self.jt])
            self.jt += 1

        self.Pause()


    def PreparePloting(self):
 # Set plot ranges and titles

        maxlim = mquantiles(self.m2g.Y*self.m2g.norm, .98)
        minlim = mquantiles(self.m2g.Y*self.m2g.norm, .02)
        if minlim == maxlim: minlim, maxlim=0, 1
        self.ax1.set_ylim(minlim-(maxlim-minlim)*.1, maxlim+(maxlim-minlim)*.3)

        maxlim = mquantiles(self.m2g.K[:-1, :], .98)
        minlim = min(0, mquantiles(self.m2g.K[:-1, :], .02))
        if minlim == maxlim: minlim, maxlim=0, 1
        self.ax2.set_ylim(minlim-(maxlim-minlim)*.1, maxlim+(maxlim-minlim)*.1)
        self.ax1.set_title('Fitted data: ' + self.ext_str + ',   '+'$\chi^2/doF$: %.2f'%self.m2g.chi2)
        self.plot_step()


    def updateTimeSlider(self, val):
        self.sl_time.val = val
        poly = self.sl_time.poly.get_xy()
        poly[2:4, 0] = val
        self.sl_time.poly.set_xy(poly)   

        
    def plot_step(self):
# Plot single step

        nx = self.xexp.shape[1]
        imin = nx*self.jt
        self.ind = range(imin, imin + nx)

        self.time_text.set_text('time: %.3fs' %self.tgrid[self.jt])

        if any(self.ind):
            self.replot_plot.set_visible(True)
            self.plotline.set_visible(True)
            for c in self.caplines:
                c.set_visible(True)
            self.barlinecols[0].set_visible(True)
            x    = self.m2g.R[self.ind]
            y    = self.m2g.Y[self.ind]*self.m2g.norm
            yerr = self.m2g.dy[self.ind]*self.m2g.norm

            ry = self.m2g.retro_f[self.ind]
            self.replot_plot.set_data(x, ry)

# Replot the data first
            self.plotline.set_data(x, y)

# Update the caplines
            self.caplines[0].set_data((x, y-yerr))
            self.caplines[1].set_data((x, y+yerr))

#Update the error bars
            self.barlinecols[0].set_segments(zip(zip(x, y - yerr), zip(x, y + yerr))) 
        else:
            self.replot_plot.set_visible(False)
            self.plotline.set_visible(False)
            for c in self.caplines:
                c.set_visible(False)
            self.barlinecols[0].set_visible(False)

        self.fit_plot.set_data(self.m2g.r_new[self.jt], self.m2g.g[self.jt])

        paths, = self.confidence_plot.get_paths()

#inv K and errorbars
        self.k_plot.set_data(self.m2g.r_new[self.jt], self.m2g.K[:, self.jt])
        
        u_err = self.m2g.K[:-1, self.jt] + self.m2g.Kerr_u[:-1, self.jt]
        b_err = self.m2g.K[:-1, self.jt] - self.m2g.Kerr_d[:-1, self.jt]

        paths.vertices[1:self.nr_fit, 1] = u_err
        paths.vertices[self.nr_fit+1:-1, 1] = b_err[::-1]
        paths.vertices[self.nr_fit, 1] = b_err[-1]
        paths.vertices[0, 1] = b_err[0]
        
        self.fig.canvas.draw()


    def plot3d(self):

        fsize3d = 16
        plt.ion()
        fig1 = plt.figure('3D plot')
        ax = p3.Axes3D(fig1)
        ax.set_xlabel(self.xlab, fontsize=fsize3d)
        ax.set_ylabel('Time [s]', fontsize=fsize3d)
        
        ax.plot_wireframe(self.m2g.r_new, self.m2g.t_new, self.m2g.g, linewidth=.3, rstride=1, cstride=3)
        plt.draw()


    def MouseInteraction(self, event):

        if event.button == 1 and self.ctrl:
            self.delete_channel(event) 
        elif event.button == 1:
            self.delete_point(event) 
        elif event.button == 2:
            self.fit()
        elif event.button == 3 and self.ctrl:
            self.stop=True
            self.Backward() 
        elif event.button == 3:
            self.stop=True
            self.Forward() 


    def delete_channel(self, event):
        self.delete_point(event, True)


    def delete_point(self, event, all_channel=False):
# Set point errorbars to infinity
        if self.ax1 != event.inaxes:
            return
        self.ind &= np.isfinite(self.m2g.dy)
        if not any(self.ind):
            logger.error('No point in the time interval')
            return

        sx = self.ax1.get_xlim()
        sx = sx[-1] - sx[0]
        sy = self.ax1.get_ylim()
        sy = sy[-1] - sy[0]

        x = self.m2g.R[self.ind]
        y = self.m2g.Y[self.ind]*self.m2g.norm
        i = np.argmin(np.hypot((x - event.xdata)/sx, (y - event.ydata)/sy))
        i_ind = np.where(self.ind)[0]

        if all_channel:
            ch = self.channel[i_ind[i]]
            self.m2g.dy[self.channel==ch] = np.infty

        self.m2g.dy[i_ind[i]]= np.infty

        self.plot_step()


    def on_key(self, event):
        if 'control' == event.key:
            self.ctrl = True
        if 'shift' == event.key:
            self.shift=True
        if ' ' == event.key:
            if self.stop:
                self.Play()
            else:
                self.Pause()


    def off_key(self, event):
        if event.key in ('ctrl+control', 'control') :
            self.ctrl=False
        if 'shift' == event.key:
            self.shift=False
