#!/usr/bin/env python

import os, sys, logging
try:
    import Tkinter as tk
    import ttk
except:
    import tkinter as tk
    from tkinter import ttk
import numpy as np
import aug_sfutils as sfu
import trview
from trview import profiles, rhop2rhot, main_spec, tkhyper, fit_sep, \
    plot_geo, plot_sig, plot_power, tooltip, \
    aug2sig, aug2ech, aug2nbi, aug2icrf, \
    write_u, write_rabbit, astra_nml_u, astra_exp, wrnml

fmt = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s: %(message)s', '%H:%M:%S')
hnd = logging.StreamHandler()
hnd.setFormatter(fmt)
logger = logging.getLogger('TRV_GUI')
logger.addHandler(hnd)
logger.setLevel(logging.INFO)

try:
    from trview import trv2imas
except:
    logger.warning('Missing IMAS output, do:')
    logger.warning('   on IPP-toki: module load imas-modules/1.0 IMAS/3.25.0-4.4.0 trview')
    logger.warning('   on gateway : module load cineca imasenv trview')

dia_init_d = {'I': ('Ti', 'CEZ:Ti_c', 'CMZ:Ti_c'), 'E': ('Te', 'IDA:Te'), 'N': ('ne', 'IDA:ne'), 'V': ('vtor', 'CEZ:vrot', 'CMZ:vrot')}

TG_INFO = '\n---------------------------\n' + \
'TRVIEW version %s\n' %(trview.__version__) + \
"""Including IMAS output
All exp data (separatrix, traces, profiles) fetched in one step
Enforced equilibrium consistency for separatrix & data mapping
TRANSP, ASTRA input created in one step (respectively)
Warning: Unlimited backup for pre-existing u-files

Author: Giovanni Tardini
git@ipp.mpg.de
---------------------------"""

def about():

    mytext = 'Documentation at the <a href="http://www.aug.ipp.mpg.de/aug/manuals/transp/trview/index.html">TRVIEW homepage</a>'
    h = tkhyper.HyperlinkMessageBox("Help", mytext, "340x60")


class GUI_PAR():

    pass


class TRVIEW:


    def __init__(self):


        logger.info(TG_INFO)

        trgui = tk.Tk()
        xmax = trgui.winfo_screenwidth()
        ymax = trgui.winfo_screenheight()
        width  = min(1200, int(0.95*xmax))
        height = min(800, int(0.88*ymax)) 
        trgui.geometry('%dx%d' %(width, height))
        trgui.title('TRVIEW')

#        gui.option_add("*Font", "Helvetica")
# Widgets style

        from trview import trgui_style

# Menubar

        menubar  = tk.Menu(trgui)
        filemenu = tk.Menu(menubar, tearoff=0)
        filemenu.add_command(label="Read/fit exp data", command=self.get_data)
        if 'imas' in sys.modules:
            filemenu.add_command(label="Fill IDS tree", command=self.fill_ids)
        filemenu.add_command(label="Write TRANSP input", command=self.wr_tr)
        filemenu.add_command(label="Write ASTRA input", command=self.wr_as)
        filemenu.add_command(label="Write RABBIT input", command=self.wr_rb)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=sys.exit)

        helpmenu = tk.Menu(menubar, tearoff=0)
        helpmenu.add_command(label="About", command=about)

        menubar.add_cascade(label = "File", menu=filemenu)
        menubar.add_cascade(label = "Help", menu=helpmenu)

        menubar.configure(bg=trgui_style.tbc)
        trgui.config(menu = menubar)

# Main frame is a notebook
        self.nb_trgui = ttk.Notebook(trgui, name='notebo')
        self.nb_trgui.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

 # General entries

        setup_frame    = ttk.Frame(self.nb_trgui)
        self.nb_trgui.add(setup_frame, text='Settings')

        shotframe = ttk.Frame(setup_frame)
        timeframe = ttk.Frame(setup_frame)
        equframe  = ttk.Frame(setup_frame)
        neframe   = ttk.Frame(setup_frame)
        teframe   = ttk.Frame(setup_frame)
        tiframe   = ttk.Frame(setup_frame)
        vtframe   = ttk.Frame(setup_frame)
        nmlframe  = ttk.Frame(setup_frame)

        for frame in shotframe, timeframe, equframe, \
            neframe, teframe, tiframe, vtframe, nmlframe:
            frame.pack(side=tk.TOP, fill=tk.BOTH, pady=6)

        self.gui_d = {}
        lbl_d = {}

# nshot, tbeg, tend
        set_init = {'shot': 39649, 'IDSrun': 6, 'nrho': 51, 'n_the': 91, 'tbeg': 0., 'tend': 10., 'dt_prof': 0.05, 'dt_sig': 0.005}
        for key in ('shot', 'IDSrun', 'nrho', 'n_the'):
            lbl_d[key] = ttk.Label(shotframe, text=key, width=8)
            lbl_d[key].pack(side=tk.LEFT, padx='10 1')
            var = tk.IntVar()
            var = ttk.Entry(shotframe, width=8)
            var.insert(0, set_init[key])
            var.pack(side=tk.LEFT)
            self.gui_d[key] = var

        for key in ('tbeg', 'tend', 'dt_prof', 'dt_sig'):
            lbl_d[key] = ttk.Label(timeframe, text=key, width=8)
            lbl_d[key].pack(side=tk.LEFT, padx='10 1')
            var = tk.DoubleVar()
            var = ttk.Entry(timeframe, width=8)
            var.insert(0, set_init[key])
            var.pack(side=tk.LEFT)
            self.gui_d[key] = var

        var = tk.BooleanVar()
        var.set(True)
        cbt = ttk.Checkbutton(timeframe, text='No ELMs', variable=var, width=10)
        cbt.pack(side=tk.LEFT, anchor=tk.W, padx='10 1')
        self.gui_d['no_elm'] = var

        tooltip.createToolTip(lbl_d['shot']   , 'AUG shot number')
        tooltip.createToolTip(lbl_d['IDSrun'] , 'Run identifier for IMAS database')
        tooltip.createToolTip(lbl_d['nrho']   , 'Size of spline rho-grid for profiles')
        tooltip.createToolTip(lbl_d['n_the']  , '#points of separatrix fit')
        tooltip.createToolTip(lbl_d['tbeg']   , 'Initial time of input collection')
        tooltip.createToolTip(lbl_d['tend']   , 'Final time of input collection')
        tooltip.createToolTip(lbl_d['dt_prof'], 'Time step of profiles time grid')
        tooltip.createToolTip(lbl_d['dt_sig'], 'Time sampling of time traces')
        tooltip.createToolTip(cbt, 'Checked: exclude time frames [-1ms, 5ms] around ELM crash times')

# Separatrix

        geoinit = {'Nmom': 6, 'eq_exp': 'AUGD', 'eq_diag': 'EQI', 'eq_ed': 0}
        for key in ('eq_exp', 'eq_diag', 'eq_ed', 'Nmom'):
            lbl_d[key] = ttk.Label(equframe, text=key, width=8)
            lbl_d[key].pack(side=tk.LEFT, padx='10 1')
            var = ttk.Entry(equframe, width=8)
            var.insert(0, geoinit[key])
            var.pack(side=tk.LEFT)
            self.gui_d[key] = var

        tooltip.createToolTip(lbl_d['Nmom'], '# of Fourier moments for separatrix fit, the higher, the more accurate/asymmetric')
        tooltip.createToolTip(lbl_d['eq_exp'], 'Equilibrium shotfile exp label')
        tooltip.createToolTip(lbl_d['eq_diag'], 'Equilibrium shotfile diag label')
        tooltip.createToolTip(lbl_d['eq_ed'], 'Equilibrium shotfile edition')

# Fit method

        self.fit_d = {}
        self.cbvar = {}
        self.ent_sflist = {}
        self.diag_choice('N', neframe)
        self.diag_choice('E', teframe)
        self.diag_choice('I', tiframe)
        self.diag_choice('V', vtframe)
        for prefix in dia_init_d.keys():
            self.set_sflist(prefix)

# TRANSP/NUBEAM namelist

        self.nml_d = {}

# Frame structure

        ttk.Label(nmlframe, text='TRANSP/NUBEAM namelist', width=30).pack(side=tk.TOP, anchor=tk.W, padx=1)
        tr_entframe = ttk.Frame(nmlframe)
        tr_avgframe = ttk.Frame(nmlframe)
        tr_fbmframe = ttk.Frame(nmlframe)
        for frame in tr_entframe, tr_avgframe, tr_fbmframe:
           frame.pack(side=tk.TOP, anchor=tk.W, pady=4)

        nml_init = {'tres': '0.01', 'dtbeam': '0.01', 'nmc': '50000', 'avgtim': 0.02, 'fbm': '0.'}

        for key in ('tres', 'dtbeam', 'nmc'):
            lbl_d[key] = ttk.Label(tr_entframe, text=key, width=7)
            lbl_d[key].pack(side=tk.LEFT, padx='10 1')
            var = ttk.Entry(tr_entframe, width=6)
            var.insert(0, nml_init[key])
            var.pack(side=tk.LEFT)
            self.gui_d[key] = var

        var = tk.BooleanVar()
        var.set(False)
        rfb = ttk.Checkbutton(tr_entframe, variable=var, text='TORIC FBM', width=10)
        rfb.pack(side=tk.LEFT, anchor=tk.W, padx=15)
        self.gui_d['mcrf'] = var

        lbl_d['avgtim'] = ttk.Label(tr_avgframe, text='AVGTIM', width=8)
        lbl_d['avgtim'].pack(side=tk.LEFT, padx=1)
        var = tk.DoubleVar()
        var = ttk.Entry(tr_avgframe, width=6)
        var.insert(0, nml_init['avgtim'])
        var.pack(side=tk.LEFT)
        self.gui_d['avgtim'] = var

        lbl_d['fbm'] = ttk.Label(tr_fbmframe, text='FBM times')
        lbl_d['fbm'].pack(side=tk.LEFT, anchor=tk.W, padx=1)
        var = ttk.Entry(tr_fbmframe, width=30)
        var.insert(0, nml_init['fbm'])
        var.pack(side=tk.LEFT)
        self.gui_d['fbm'] = var

        tooltip.createToolTip(lbl_d['nmc'], '# MonteCarlo markers')
        tooltip.createToolTip(lbl_d['tres'], 'Output sampling [s]')
        tooltip.createToolTip(lbl_d['dtbeam'], 'Time between NUBEAM calls [s]')
        tooltip.createToolTip(rfb, 'Checked: switch on RF-kick operator')
        tooltip.createToolTip(lbl_d['avgtim'], 'Time interval for FBM distribution, left of each OUTTIM(j)')
        tooltip.createToolTip(lbl_d['fbm'], 'List of FBM (OUTTIM) times, space-separated')

# Display GUI
        trgui.mainloop()


    def diag_choice(self, prefix, diagframe):

        lblframe = ttk.Frame(diagframe)
        diaframe = ttk.Frame(diagframe)
        lblframe.pack(side=tk.TOP, anchor=tk.W, expand=1)
        diaframe.pack(side=tk.TOP, anchor=tk.W, expand=1)

        ttk.Label(lblframe, text=dia_init_d[prefix][0], width=10).pack(side=tk.LEFT)
        var = tk.StringVar()
        var.set('rec_spl')
        fit_rb1 = ttk.Radiobutton(lblframe, text='Rec. spline', variable=var, value='rec_spl')
        fit_rb2 = ttk.Radiobutton(lblframe, text='Gauss', variable=var, value='gauss')
        fit_rb3 = ttk.Radiobutton(lblframe, text='Fit 2D', variable=var, value='fit_2d')
        self.fit_d[prefix] = var
        for rb in fit_rb1, fit_rb2, fit_rb3:
            rb.pack(side=tk.LEFT, padx=10)
        if prefix == 'V':
            var = tk.BooleanVar()
            var.set(True)
            ang_cb = ttk.Checkbutton(lblframe, text='Ang.freq', variable=var)
            ang_cb.pack(side=tk.RIGHT, anchor=tk.E, padx='100 1')
            self.gui_d['angf'] = var
        tooltip.createToolTip(fit_rb1, 'Cubic spline of single profiles')
        tooltip.createToolTip(fit_rb2, 'Gaussian fit of single profiles')
        tooltip.createToolTip(fit_rb3, 'Fast 2d spline fitting of the profiles')

        self.cbvar[prefix] = {}
        jcol = 0

# Browse all diagnostics with a given prefix

        f_diags = '%s/diags.dat' %os.path.dirname(os.path.realpath(__file__))
        diags, sigs, pre = \
            np.genfromtxt(f_diags, dtype='str', skip_header=2, unpack=True, usecols=(0, 1, 2))

        for jdia, diag in enumerate(diags):
            if pre[jdia] == prefix:
                sig = sigs[jdia]
                diag_sig = '%s:%s' %(diag, sig)
                self.cbvar[prefix][diag_sig] = tk.BooleanVar()
                self.cbvar[prefix][diag_sig].set(False)
                if (diag_sig in dia_init_d[prefix][1:]):
                    self.cbvar[prefix][diag_sig].set(True)
                cbt = ttk.Checkbutton(diaframe, text=diag_sig, variable=self.cbvar[prefix][diag_sig], command=lambda x=prefix: self.set_sflist(x), width=10)
                cbt.pack(side=tk.LEFT, fill=tk.X, expand=1)
                if jcol%8 == 7:
                    diaframe = ttk.Frame(diagframe)
                    diaframe.pack(side=tk.TOP, anchor=tk.W)
                jcol += 1

# Summary with diags
        listframe = ttk.Frame(diagframe)
        listframe.pack(side=tk.TOP, anchor=tk.W)
        var = tk.StringVar()
        var = ttk.Entry(listframe, width=105)
        var.insert(0, '')
        var.pack(side=tk.LEFT)
        self.ent_sflist[prefix] = var      


    def get_gui(self):

        gpar = GUI_PAR()
        for key in self.gui_d.keys():
            stmp = self.gui_d[key].get()
            if key in ('shot', 'IDSrun', 'eq_ed', 'Nmom', 'nrho', 'n_the', 'nmc'):
                gpar.__dict__[key] = int(stmp)
            elif key in ('tbeg', 'tend', 'dt_prof', 'dt_sig', 'tres', 'dtbeam', 'avgtim'):
                gpar.__dict__[key] = float(stmp)
            elif key in ('no_elm', 'mcrf', 'angf'):
                gpar.__dict__[key] = stmp
            else:
                gpar.__dict__[key] = stmp.strip()

        outtim = np.sort([float(x) for x in self.gui_d['fbm'].get().split()])
        ind = (outtim > gpar.tbeg + gpar.avgtim) & (outtim <= gpar.tend)
        gpar.outtim = outtim[ind]

        return gpar


    def set_sflist(self, prefix):

        sf_list = [sf for sf, chosen in self.cbvar[prefix].items() if chosen.get()]
        sflist = ''
        for jsig, dsig in enumerate(sf_list):
            tmp = 'AUGD:%s:0' %dsig
            sflist += '%s ' %tmp 

        self.ent_sflist[prefix].delete(0, tk.END)
        self.ent_sflist[prefix].insert(0, sflist)


    def get_data(self):

        gpar = self.get_gui()

        ion_spec = main_spec.spec_jou_sf(gpar.shot)
        logger.info('Main ion species is %s', ion_spec)
        self.ion_d = {'Aimp': 12, 'Zimp': 6}
        if ion_spec == 'D':
            self.ion_d['Amain'] = 2
            self.ion_d['Zmain'] = 1
        elif ion_spec == 'H':
            self.ion_d['Amain'] = 1 
            self.ion_d['Zmain'] = 1
        elif ion_spec == 'He':
            self.ion_d['Amain'] = 4 
            self.ion_d['Zmain'] = 2 

        self.nbi = aug2nbi.NBI  (gpar.shot, tbeg=gpar.tbeg, tend=gpar.tend, dt_ds=gpar.dt_sig)
        if self.nbi.status:
            self.nbi.aug_tr()
        self.ech = aug2ech.gyros(gpar.shot, tbeg=gpar.tbeg, tend=gpar.tend, dt_ds=gpar.dt_sig)
        self.icr = aug2icrf.ICRF(gpar.shot, tbeg=gpar.tbeg, tend=gpar.tend, dt_ds=gpar.dt_sig)
        self.equ = sfu.EQU(gpar.shot, tbeg=gpar.tbeg, tend=gpar.tend, diag=gpar.eq_diag, exp=gpar.eq_exp, ed=gpar.eq_ed)
        gpar.eq_ed = self.equ.ed
        self.equ.geofit = fit_sep.geofit(self.equ, nmom=gpar.Nmom, n_the=gpar.n_the)
        self.sig = aug2sig.READ(self.equ, dt_ds=gpar.dt_sig)

        self.plot_sepx()
        self.plot_traces()
        sflists = {}
        for prefix, entry in self.ent_sflist.items():
            sflists[prefix] = entry.get().strip().split()
        for key in ('ti', 'te', 'ne', 'vrot'): 
            if key in self.nb_trgui.children.keys():
                self.nb_trgui.children[key].destroy()
        obj = profiles.EXP_PROF(gpar, self.equ, self.nb_trgui, sflists, self.fit_d)
        self.profiles = obj.pr_obj
        self.ext_d = {}
        for prefix, sfls in obj.sfl_d.items():
            if prefix not in obj.pr_obj.prof.keys():
                continue
            if prefix == 'V':
                if obj.pr_obj.prof['V'].units == 'rad/s':
                    prefix = 'ANGF'
            self.ext_d[prefix] = ''
            for sfl in sfls:
                diag = sfl.split(':')[1]
                self.ext_d[prefix] += diag


    def plot_sepx(self):

        if 'sep' in self.nb_trgui.children.keys():
            self.nb_trgui.children['sep'].destroy()
        sepframe = ttk.Frame(self.nb_trgui, name='sep')
        self.nb_trgui.add(sepframe , text='Sep fit')
        plot_geo.GEO(self.equ.geofit, geoframe=sepframe)


    def plot_traces(self):

        if 'traces' in self.nb_trgui.children.keys():
            self.nb_trgui.children['traces'].destroy()
        traceframe = ttk.Frame(self.nb_trgui, name='traces')
        self.nb_trgui.add(traceframe, text='Traces')

        nbout = ttk.Notebook(traceframe, name='plot output')
        nbout.pack(side=tk.TOP, fill=tk.BOTH)
        sigframe = ttk.Frame(nbout)
        nbiframe = ttk.Frame(nbout)
        ecrframe = ttk.Frame(nbout)
        icrframe = ttk.Frame(nbout)
        nbout.add(sigframe, text='Time traces')
        plot_sig.plot_sig(self.sig, sigframe=sigframe)
        if hasattr(self.nbi, 't_nis'):
            plot_power.plot_power(self.nbi.t_ds, self.nbi.pnbi_ds, self.nbi.shot, powframe=nbiframe)
            nbout.add(nbiframe, text='P_NBI')
        if hasattr(self.ech, 't_ds'):
            plot_power.plot_power(self.ech.t_ds, self.ech.pech_ds, self.ech.shot, powframe=ecrframe)
            nbout.add(ecrframe, text='P_ECRH')
        if hasattr(self.icr, 't_icp'):
            plot_power.plot_power(self.icr.t_ds, self.icr.picr_ds, self.icr.shot, powframe=icrframe)
            nbout.add(icrframe, text='P_ICRF')


    def wr_tr(self):

        if not hasattr(self, 'nbi'):
            logger.error('Fetch exp data first')
            return

        write_u.tr_u_geo(self.equ.geofit)
        write_u.write_u1d(self.sig, self.nbi, self.ech, self.icr)
        rhop2rhot.rhop2rhot(self.equ, self.profiles)
        write_u.write_q(self.equ)
        write_u.write_u2d(self.profiles)
        gpar = self.get_gui()
        runid = wrnml.w_nml(gpar, self.ion_d, self.ext_d, self.nbi, ech=self.ech, icr=self.icr)


    def wr_as(self):

        if not hasattr(self, 'nbi'):
            logger.error('Fetch exp data first')
            return

        rhop2rhot.rhop2rhot(self.equ, self.profiles)
        write_u.write_u2d(self.profiles)
        astra_nml_u.astra_nml_u(self.equ.geofit, self.sig, self.nbi, self.ech)
        astra_exp.astra_exp(self.equ.shot, self.ion_d, self.ext_d, self.nbi)


    def wr_rb(self):

        if not hasattr(self, 'nbi'):
            logger.error('Fetch exp data first')
            return

        rhop2rhot.rhop2rhot(self.equ, self.profiles)
# Omega, not vtor

        write_rabbit.write_rabbit(self.profiles, self.equ, self.sig.zef, self.nbi, self.ion_d)


    def fill_ids(self):

        if not hasattr(self, 'nbi'):
            logger.error('Fetch exp data first')
            return

        if 'imas' not in sys.modules:
            logger.error('No IMAS available')
            return
        gpar = self.get_gui()

# Create IMAS object, initialise IDS structure
        trv2imas.TRV2IMAS(gpar, self.ion_d, self.equ, self.profiles, self.sig, \
            self.nbi, self.ech, self.icr, write_ids=True)

if __name__ == "__main__":


    TRVIEW()
