import os, logging
try:
    import Tkinter as tk
    import ttk
except:
    import tkinter as tk
    from tkinter import ttk
import numpy as np
from scipy.interpolate import interp1d

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
try:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk as nt2tk
except:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg as nt2tk
from matplotlib.figure import Figure
import mpl_toolkits.mplot3d.axes3d as p3
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.widgets import Slider
from multiprocessing import Pool, cpu_count
from trview import rec_spline1d, gauss
from trview.tooltip import createToolTip

logger = logging.getLogger('trview.vprof')
logger.setLevel(logging.INFO)
#logger.setLevel(logging.DEBUG)


fsize = 12
fsize3d = 16


def fit1d(tup_in):

    x, y, err, rho, time, fit_tol, sigmas, method = tup_in
    nx = len(x)
    nfit = len(rho)
    if np.isnan(y).all():
        logger.debug('fit1d, nans only at t=%.3f s', time) 
        if method == 'gauss':
            return np.zeros(nfit + nx + nfit)
        else:
            return np.zeros(nfit + nx)

    drho = rho[1] - rho[0]
    rho_ext = rho
    while rho_ext[-1] < x[-1]:
        rho_ext = np.append(rho_ext, rho_ext[-1] + drho)
    if np.max(np.abs(y)) == 0:
        if method == 'gauss':
            return np.zeros(nfit + nx + nfit)
        else:
            return np.zeros(nfit + nx)

# Discard zeros
    indzero = (y == 0)
    y[indzero]   = np.nan
    err[indzero] = np.infty
    if fit_tol == 0:
        ind = np.isfinite(y)
        yfit = np.interp(rho, x[ind], y[ind])
        if method == 'gauss':
            confid = np.zeros(nfit)
    else:
        if method == 'rec_spl':
            try:
                yfit_ext, y = rec_spline1d.spline_rec1d(x, y, rho_ext, \
                     y_err=err, fit_tol=fit_tol, sigmas=sigmas)
            except:
                logger.error('Failed rec_spline')
                yfit_ext = np.zeros(nfit)
        elif method == 'gauss':
            ind = np.isfinite(err)
            yfit_ext, confid = gauss.gauss(x[ind], y[ind], err[ind], rho_ext, fit_tol=fit_tol)
#            yfit_ext, confid = gauss.gauss(x, y, err, rho_ext, fit_tol=fit_tol)
            confid = confid[:nfit]
        yfit = yfit_ext[:nfit]

    out = np.append(yfit, y)
    if method == 'gauss':
        out = np.append(out, confid)

    return out


class MPLWID:


    timeout = 1e-10


    def __init__(self, nshot, n_spl, red_d, myframe=None, method='rec_spl'):


        self.method = method
        self.shot   = nshot

# Choosing last diag's info for plotting
        self.xlab = r'$\rho_{pol}$'

# Combine diagnostics

        for data_d in red_d.values():
            if not hasattr(self, 'xexp'):
                self.xexp = np.array(data_d['rhop'])
                self.yexp = np.array(data_d['data'])
                self.yerr = np.array(data_d['data_err'])
            else:
                self.xexp = np.append(self.xexp, data_d['rhop']    , axis=1)
                self.yexp = np.append(self.yexp, data_d['data']    , axis=1)
                self.yerr = np.append(self.yerr, data_d['data_err'], axis=1)

        ind_sort = np.argsort(self.xexp, axis=1)
        for jt in range(self.xexp.shape[0]):
            self.xexp[jt,:] = self.xexp[jt, ind_sort[jt]]
            self.yexp[jt,:] = self.yexp[jt, ind_sort[jt]]
            self.yerr[jt,:] = self.yerr[jt, ind_sort[jt]]

        self.tgrid = data_d['tgrid']

        self.jt = 0
        self.nt = len(self.tgrid)
        self.rho_spl = np.linspace(0, 1, n_spl)
        self.xg, self.tg = np.meshgrid(self.rho_spl, self.tgrid)

# Frame structure

        if myframe is None:
            myframe = tk.Toplevel()
            myframe.geometry('1200x740')
            myframe.title('View and fit profiles, method %s' %self.method)

        figframe  = ttk.Frame(myframe)
        entframe  = ttk.Frame(myframe)
        lblframe  = ttk.Frame(myframe)
        toolframe = ttk.Frame(myframe)
        figframe.pack(side=tk.TOP, fill=tk.BOTH, expand=1) 
        for box in (toolframe, entframe, \
                    lblframe):
            box.pack(side=tk.TOP, fill=tk.X)

        figframe3 = ttk.Frame(figframe)
        figframe2 = ttk.Frame(figframe)
        for box in (figframe3, figframe2):
            box.pack(side=tk.LEFT, fill=tk.X)

        figframe1 = ttk.Frame(figframe3)
        entfframe = ttk.Frame(figframe3)
        for box in (figframe1, entfframe):
            box.pack(side=tk.TOP, fill=tk.X)

        self.nb2d = ttk.Notebook(figframe2)
        self.nb2d.pack(side=tk.TOP, fill=tk.BOTH, expand=1)
        frame2d = ttk.Frame(self.nb2d)
        ctframe = ttk.Frame(self.nb2d)
        self.nb2d.add(frame2d, text='Wired plot')
        self.nb2d.add(ctframe, text='Contour plot')

# Plot settings

        self.fig1d = Figure(figsize=(5., 4.3), dpi=100)
        self.fig2d = Figure(figsize=(5.6, 4.5), dpi=100)
        self.figct = Figure(figsize=(5.6, 4.5), dpi=100)

        can_1d = FigureCanvasTkAgg(self.fig1d, master=figframe1)
        can_2d = FigureCanvasTkAgg(self.fig2d, master=frame2d)
        can_ct = FigureCanvasTkAgg(self.figct, master=ctframe)
        for canv in can_1d, can_2d, can_ct:
            canv._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        toolbar = nt2tk( can_1d, toolframe)
        toolbar.update()

        ylab = ''
        for sflist in sorted(red_d.keys()):
            diag = sflist.split(':')[1]
            ylab += ',%s' %diag
        ylab = ylab[1:]
        ylab += ' [%s]' %red_d[sflist]['unit']

        self.ax1d = self.fig1d.add_subplot(1, 1, 1)
        self.fig1d.subplots_adjust(left=0.15, bottom=0.16, right=0.95, top=0.9)

        self.ax1d.set_xlim([0, 1.1])
        self.ax1d.set_xlabel(self.xlab, fontsize=fsize)
        self.ax1d.set_ylabel(ylab, fontsize=fsize)

        self.ok, self.caplines, self.barlinecols = \
            self.ax1d.errorbar(0, 0, 0, fmt='g.', capsize=3.)
        self.rej,    = self.ax1d.plot([], [], 'ro')
        self.del_pt, = self.ax1d.plot([], [], 'bo')
        self.pfit,   = self.ax1d.plot([], [], 'm-')
        self.sep_line, = self.ax1d.plot([1, 1], [0, 1], 'k--')

# Axes of 2d figure

        self.ax2d = p3.Axes3D(self.fig2d)

# Player buttons

        locdir = os.path.dirname(os.path.realpath(__file__))
        self.playfig  = tk.PhotoImage(file='%s/play.gif'     %locdir)
        self.pausefig = tk.PhotoImage(file='%s/pause.gif'    %locdir)
        self.forwardfig    = tk.PhotoImage(file='%s/forward.gif'  %locdir)
        self.backwardfig   = tk.PhotoImage(file='%s/backward.gif' %locdir)

        self.backward_button = ttk.Button(entfframe, command=self.Backward, image=self.backwardfig)
        self.play_button     = ttk.Button(entfframe, command=self.Play,     image=self.playfig)
        self.forward_button  = ttk.Button(entfframe, command=self.Forward,  image=self.forwardfig)

        createToolTip(self.forward_button  , 'Go to next timestep')
        createToolTip(self.backward_button , 'Go backward by one timestep')
        createToolTip(self.play_button     , 'Forward animation/pause')
        try: # matplotlib > 2.0
            sliderax = self.fig1d.add_axes([0.2, 0.02, 0.6, 0.03], facecolor='yellow')
        except:
            sliderax = self.fig1d.add_axes([0.2, 0.02, 0.6, 0.03], axisbg='yellow')
        self.sl_time = Slider(sliderax,'Time:', 0., 1., valinit=0.)
        self.sl_time.valtext.set_text('0.0000')

        def update_time(val):
            self.update_plot(val=val)

        self.sl_time.on_changed(update_time)
        self.fig1d.canvas.mpl_connect('button_press_event', self.MouseInteraction)
        self.fig1d.canvas.mpl_connect('key_press_event'   , self.on_key)
        self.fig1d.canvas.mpl_connect('key_release_event' , self.off_key)
        self.fig1d.canvas.mpl_connect('button_press_event', lambda event:self.fig1d.canvas._tkcanvas.focus_set())

        for but in self.backward_button, self.play_button, self.forward_button:
            but.pack(side=tk.LEFT)

# Fit entries
        btn_fit = ttk.Button(entframe, text='Re-fit', command=self.fit)
        btn_fit.pack(side=tk.LEFT, padx=2)
        sigma_init = 1
        logger.debug(ylab)
        if ylab[:3] == 'IDA':
            erbar_init = 0.
        else:
            erbar_init = .3

        lbl_fit = ttk.Label(entframe, text='Fit tolerance', width=11)
        lbl_fit.pack(side=tk.LEFT, padx='10 1')
        self.cfit_val = ttk.Entry(entframe, width=8)
        self.cfit_val.insert(0, erbar_init)
        self.cfit_val.pack(side=tk.LEFT, padx=2)

        createToolTip(btn_fit, 'Re-fit data with current fit tolerance, sigmas')
        createToolTip(lbl_fit, 'Multiplier for error bars, the higher, the smoother the fit')

        if self.method == 'rec_spl':
            lbl_sig = ttk.Label(entframe, text='# sigmas')
            lbl_sig.pack(side=tk.LEFT, padx='10 1')
            self.sigma_val = ttk.Entry(entframe, width=8)
            self.sigma_val.insert(0, sigma_init)
            self.sigma_val.pack(side=tk.LEFT, padx=2)
            createToolTip(lbl_sig, '# sigmas, out of which scatter data are deleted')

        lblframe1 = ttk.Frame(lblframe)
        lblframe2 = ttk.Frame(lblframe)
        lblframe1.pack(side=tk.LEFT)
        lblframe2.pack(side=tk.RIGHT)

        scat_help  = tk.Label(lblframe1, text='Left plot, scatter points:')
        scat_blue  = tk.Label(lblframe1, text='Blue: hand-deleted', fg='blue')
        scat_red   = tk.Label(lblframe1, text='Red: rejected by spline', fg='red')
        scat_green = tk.Label(lblframe1, text='Green: valid points', fg='#00a000')
        for w in (scat_help, scat_blue, scat_red, scat_green):
            w.pack(side=tk.TOP, anchor=tk.W)

        del_lb  = ttk.Label(lblframe2, text='Mouse-L: deletes points')
        del_all = ttk.Label(lblframe2, text='Ctrl+Mouse-L: deletes frame')
        undolb  = ttk.Label(lblframe2, text='Mouse-Mid: un-delete points for this frame')
        righlb  = ttk.Label(lblframe2, text='Mouse-R: move to next frame plot')
        for w in (del_lb, del_all, undolb, righlb):
            w.pack(side=tk.TOP, anchor=tk.W)

# Create GUI

        self.stop  = True
        self.ctrl  = False
        self.fit()


    def fit(self):

# Input ready, start execution

        self.ind_del  = [ [] for jt in range(self.nt) ]
        self.tnot = []

        logger.info('Fitting experimental profiles')

        self.fit_tol  = float(self.cfit_val.get())
        if self.method == 'rec_spl':
            self.sigmas = float(self.sigma_val.get())
            timeout_pool = 25
        else:
            self.sigmas = 1e12
            timeout_pool = 300
        if self.fit_tol == 0:
            self.sigmas = 1e12

        self.del2d = {}

# Fit 2D

        nt, nx_in = self.xexp.shape
        nx_fit = len(self.rho_spl)
        if self.method in ('rec_spl', 'gauss'):
            try:
                pool = Pool(cpu_count())
                out = pool.map_async(fit1d, [(self.xexp[jt], self.yexp[jt], self.yerr[jt], self.rho_spl, self.tgrid[jt], self.fit_tol, self.sigmas, self.method) for jt in range(nt)]).get(timeout_pool)
                out2 = np.array(out)
                self.yfit_rhop  = out2[:, :nx_fit]
                self.dat_ok = out2[:, nx_fit: nx_fit+nx_in]
                if self.method == 'gauss':
                    self.confid = out2[:, nx_fit+nx_in:]
                pool.close()
                pool.join()
            except:
                if hasattr(self, 'yfit_rhop'):
                    delattr(self, 'yfit_rhop')
                self.dat_ok = self.yexp
                logger.error('Timeout error: no fit!')

        if len(self.tgrid) > 1 and hasattr(self, 'yfit_rhop'):
# Contour plot

            self.figct.clf()
            axct = self.figct.add_subplot(1, 1, 1)
            axct.set_xlabel('Time [s]', fontsize=fsize3d)
            axct.set_ylabel(self.xlab , fontsize=fsize3d)
            ctr = axct.contourf(self.tg, self.xg, self.yfit_rhop)
            self.figct.colorbar(ctr, aspect=10, shrink=0.9)
            self.line, = axct.plot([], [], 'k-')

# 3D plot

            self.fig2d.clf()
            self.ax2d = p3.Axes3D(self.fig2d)
            self.ax2d.set_xlabel(self.xlab , fontsize=fsize3d)
            self.ax2d.set_ylabel('Time [s]', fontsize=fsize3d)

            self.ax2d.plot_wireframe(self.xg, self.tg, self.yfit_rhop, linewidth=.3, rstride=1, cstride=3)
            self.sing2d, = self.ax2d.plot([], [], [], 'r-')

        self.update_plot()

#------------
# Interaction
#------------

    def MouseInteraction(self, event):

        if event.button == 1 and self.ctrl:
            self.delete_frame(event) 
        elif event.button == 1:
            self.delete_point(event) 
        elif event.button == 2: # press mouse-middle
            self.undo(event)
        elif event.button == 3 and self.ctrl:
            self.stop = True
            self.Backward()
        elif event.button == 3:
            self.stop = True
            self.Forward() 


    def undo(self, event):

        self.ind_del[self.jt] = []
        out = fit1d((self.xexp[self.jt], self.yexp[self.jt], self.yerr[self.jt], \
            self.rho_spl, self.tgrid[self.jt], self.fit_tol, self.sigmas, self.method ))
        nx_fit = len(self.rho_spl)
        nx_in = self.xexp.shape[1]
        self.yfit_rhop[self.jt] = out[: nx_fit]
        self.dat_ok[self.jt] = out[nx_fit: nx_fit+nx_in]
        if self.method == 'gauss':
            self.confid[self.jt] = out[nx_fit+nx_in: ]

        if self.jt in self.tnot:
            self.tnot.remove(self.jt)
            self.del2d[self.jt].set_data([], [])
            self.del2d[self.jt].set_3d_properties([])
        self.update_plot()


    def del_frame(self, jt):

        if jt not in self.tnot:
            self.tnot.append(jt)
        self.ind_del[jt] = range(len(self.xexp[jt,:]))
        self.del2d[jt], = self.ax2d.plot(self.rho_spl, self.tg[jt], self.yfit_rhop[jt, :], 'b-')

        self.update_plot()


    def delete_frame(self, event):

        if self.ax1d != event.inaxes:
            return
        if not self.stop:
            return
        self.del_frame(self.jt)


    def delete_point(self, event):

        if self.ax1d != event.inaxes:
            return

        if not self.stop:
            return

        sx = self.ax1d.get_xlim()
        dx = sx[-1] - sx[0]
        sy = self.ax1d.get_ylim()
        dy = sy[-1] - sy[0]

        x = np.zeros(self.yexp.shape[1])
        y = np.zeros(self.yexp.shape[1])
        yerr = np.zeros(self.yexp.shape[1])

        x[:] = self.xexp[self.jt, :]
        y[:] = self.yexp[self.jt, :]
        yerr[:] = self.yerr[self.jt, :]

        y[self.ind_del[self.jt]] = np.nan
        yerr[self.ind_del[self.jt]] = np.infty

        idel = np.nanargmin(np.hypot((x - event.xdata)/dx, (y - event.ydata)/dy))
        self.ind_del[self.jt].append(idel)

        y[idel] = np.nan
        if np.sum(np.isfinite(y)) > 1:
            out = fit1d( (x, y, yerr, self.rho_spl, self.tgrid[self.jt], self.fit_tol, self.sigmas, self.method) )
            nx_fit = len(self.rho_spl)
            nx_in = self.xexp.shape[1]
            self.yfit_rhop[self.jt] = out[: nx_fit]
            self.dat_ok[self.jt] = out[nx_fit: nx_fit+nx_in]
            if self.method == 'gauss':
                self.confid[self.jt] = out[nx_fit+nx_in: ]
            self.update_plot()
        else:
            self.del_frame(self.jt)


    def on_key(self, event):
        if 'control' == event.key:
            self.ctrl = True
        if ' ' == event.key:
            if self.stop:
                self.Play()
            else:
                self.Pause()


    def off_key(self, event):
        if event.key in ('ctrl+control', 'control') :
            self.ctrl = False

#------------
# Animation
#------------

    def Pause(self):
        self.stop = True
        self.play_button['image']   = self.playfig
        self.play_button['command'] = self.Play

    def Forward(self):
        self.stop = True
        self.play_button['image']   = self.playfig
        self.play_button['command'] = self.Play
        self.jt += 1
        self.jt = self.jt%self.nt
        self.update_plot()

    def Backward(self):
        self.stop = True
        self.play_button['image']   = self.playfig
        self.play_button['command'] = self.Play
        self.jt -= 1
        self.jt = self.jt%self.nt
        self.update_plot()

    def Play(self):
        self.stop = False
        self.play_button['image']   = self.pausefig
        self.play_button['command'] = self.Pause
        while self.jt < self.nt-1 and not self.stop:
            self.jt += 1
            self.update_plot()
            self.fig1d.canvas.start_event_loop(self.timeout)


    def updateTimeSlider(self, val=None):
        if val is None:
            progress = (self.tgrid[self.jt] - self.tgrid[0])/(self.tgrid[-1] - self.tgrid[0])
        else:
            progress = val
        self.sl_time.val = progress
        poly = self.sl_time.poly.get_xy()
        poly[2:4, 0] = progress
        self.sl_time.poly.set_xy(poly)
        self.sl_time.valtext.set_text('%8.4f' %self.tgrid[self.jt])


    def update_plot(self, val=None):

        self.update_plot1d(val=val)
        if hasattr(self, 'yfit_rhop') and len(self.tgrid) > 1:
            self.update_plot2d(val=val)


    def update_plot1d(self, val=None):

        if val is not None:
            t_arr = self.tgrid[0] + val*(self.tgrid[-1] - self.tgrid[0])
            self.jt = np.argmin(np.abs(self.tgrid - t_arr))
        rho = self.xexp[self.jt]
        dat = self.yexp[self.jt]
        self.tlab = '#%d    Time = %7.4f [s]' %(self.shot, self.tgrid[self.jt])
        self.ax1d.set_title(self.tlab, fontsize=fsize)

        if hasattr(self, 'yfit_rhop'):
            yfit = self.yfit_rhop[self.jt]
            ymax = 1.01*max(np.nanmax(yfit), np.nanmax(dat))
            xfit = self.rho_spl
            self.pfit.set_data(xfit, yfit)
            if hasattr(self, 'confid'):
                sigma = self.confid[self.jt]
                if np.max(sigma) > 0:
                    if hasattr(self, 'fill'):
                        self.fill.remove()

                    self.fill = self.ax1d.fill_between(xfit, yfit - sigma, yfit + sigma, facecolor='r', edgecolor='None', alpha=0.1)

            rho_del = rho[self.ind_del[self.jt]]
            dat_del = dat[self.ind_del[self.jt]]

            self.del_pt.set_data(rho_del , dat_del )
            ind_ok = (np.isfinite(self.dat_ok[self.jt]))
        else:
            ind_ok = np.ones(len(rho), dtype=bool)
            ymax = 1.01*np.max(dat)

        x = rho[ind_ok]
        y = dat[ind_ok]
        err = self.yerr[self.jt, ind_ok]
        self.ok.set_data( x, y)
        self.rej.set_data(rho[~ind_ok], dat[~ind_ok])

# Update the caplines
        self.caplines[0].set_data((x, y-err))
        self.caplines[1].set_data((x, y+err))

#Update the error bars
        self.barlinecols[0].set_segments(zip(zip(x, y - err), zip(x, y + err)))

        ymin = min(np.nanmin(dat), 0)

        if np.isfinite(ymin) and np.isfinite(ymax):
            self.ax1d.set_ylim([ymin, ymax])
            self.sep_line.set_ydata([ymin, ymax])

        self.updateTimeSlider(val=val)

        self.fig1d.canvas.draw()


    def update_plot2d(self, val=None):

        if val is not None:
            t_arr = self.tgrid[0] + val*(self.tgrid[-1] - self.tgrid[0])
            self.jt = np.argmin(np.abs(self.tgrid - t_arr))
        if self.nb2d.index('current') == 0:
            self.sing2d.set_data(self.rho_spl, self.tg[self.jt])
            self.sing2d.set_3d_properties(self.yfit_rhop[self.jt])
            self.fig2d.canvas.draw()
        else:
            tim = self.tgrid[self.jt]
            self.line.set_data([tim, tim], [self.rho_spl[0], self.rho_spl[-1]])
            self.figct.canvas.draw()
